def PLTEC_initCheck() -> None:
    import os
    import json

    DEF_lang = {
        "ELF": "",
        "x86": "",
        "SET": {
            "#include:/*DVALUE/INCLUDEINS/S ": "includeFuncs",
            "BasicIO/INCLUDE": {
                "print:/*DVALUE/*DNAME/S ": "printINS",
                "input:/*DNAME/*DSIZE/S ": "inputINS",
                "clean:/S": "cleanupINS",
                "return:/*DVALUE/S ": "returnINS"
            }
        }
    }

    DEF_LanguageSET = {
        "elf": {
            "x86": {
                "funcs": {
                    "BasicIO": [
                        "print:\nmov eax, 4\nmov ebx, 1\nint 80h\n", 
                        "input:\nmov eax, 3\nmov ebx, 1\nint 80h\n", 
                        "return:\nmov eax, 1\nint 80h\n",
                        "cleanup32:\nxor eax, eax\nxor ebx, ebx\nxor ecx, ecx\nxor edx, edx\nxor esi, esi\nxor ebp, ebp\nret\n"
                    ]
                },
                "INSTRUCTION_SET": {
                    "printINS": "mov ecx, *%*DNAME*%*_pri\nmov edx, *%*DNAME*%*_pri_LEN\ncall print\n",
                    "inputINS": "mov ecx, *%*DNAME*%*_inp\nmov edx, *%*DNAME*%*_inp_LEN\ncall input\n",
                    "returnINS": "mov ebx, *%*DVALUE*%*\ncall return\n",
                    "cleanupINS": "call cleanup32\n"
                },
                "DATA_SET": {
                    "printINS": "*%*DNAME*%*_pri db \"*%*DVALUE*%*\"\n*%*DNAME*%*_pri_LEN equ $ - *%*DNAME*%*_pri\n",
                    "inputINS": "*%*DNAME*%*_inp_LEN equ *%*DSIZE*%*\n"
                },
                "BSS_SET": {
                    "inputINS": "*%*DNAME*%*_inp resb *%*DSIZE*%*\n"
                },
                "S_TEXT_END": ""
            }
        }
    }

    PLTEC = os.path.dirname(os.path.abspath(__file__))

    DL_path = os.path.join(PLTEC, "DEF_lang.json")
    DLS_path = os.path.join(PLTEC, "DEF_LanguageSET.json")

    if not os.path.exists(DL_path):
        with open(DL_path, 'w') as f:
            json.dump(DEF_lang, f)
            f.close()

    if not os.path.exists(DLS_path):
        with open(DLS_path, 'w') as f:
            json.dump(DEF_LanguageSET, f)
            f.close()


import psutil
from typing import *

import os

from psutil import *
import subprocess
import platform
import socket

SYSTEM = platform.system().lower()

class CPU:
    @staticmethod
    def CpuUsage(interval:Union[None, float]=1.0) -> float:
        """This function calculates the CPU usage via the given interval and returns the CPU usage.

        Args:
            interval (Union[None, float], optional): Interval defines time used to calculate the CPU usage. None defines the time from module import. Defaults to 1.0.

        Returns:
            float: The Cpu usage.
        """
        return psutil.cpu_percent(interval, False)
    
    @staticmethod
    def CpuUsageDetails():
        """This function returns the following details about the CPU usage ->

        user: This represents the time spent by normal processes executing in the user mode.
        system: This represents the time spent by processes executing in the kernel mode.
        idle: This represents the time when the system is idle.
        nice: This represents the time spent by priority processes executing in the user mode.
        iowait: This represents the time spent waiting for I/O to complete.
        irq: This represents the time spent for servicing hardware interrupts.
        softirq: This represents the time spent for servicing software interrupts.
        steal: Represents the time spent by other operating systems running in a virtualized environment
        guest: This represents the time spent running a virtual CPU for guest operating systems under the control of the Linux kernel.

        WINDOWS ONLY:
        interrupt: This represents the time spent for servicing hardware interrupts.
        dpc: This represents the time spent servicing deferred procedure calls (DPCs).

        Returns:
            Any: Return type -> psutil._ps[Your OS].scputimes
        """
        return psutil.cpu_times(False)
    
    @staticmethod
    def CpuUsageTimesDetails(interval:Union[float, None]=1.0):
        """Same as CpuUsage but returns like CpuUsageDetails.

        Args:
            interval (Union[None, float], optional): Interval defines time used to calculate the CPU usage. None defines the time from module import. Defaults to 1.0.

        Returns:
            Any: Return Type -> psutil._ps[Your OS].scputimes.
        """
        return psutil.cpu_times_percent(interval, False)
    
    @staticmethod
    def CpuCount(logical:bool=False) -> int:
        """Returns the number of Cpu Cores in System.

        Args:
            logical (bool, optional): If set to true returned value will be [Number of Physical Cores + Hyper Threading]. Defaults to False.

        Returns:
            int: The number of Cpu Physical Cores [+ Hyper Threading if [logical] is True].
        """
        return psutil.cpu_count(logical)
    
    @staticmethod
    def CpuStats():
        """Returns detailed CPU statistics such as context switches, interrupts, and soft interrupts.

        Returns:
            Any: Object containing detailed statistics about the CPU.
        """
        return psutil.cpu_stats()

    @staticmethod
    def CpuFreq():
        """Returns the current frequency of the CPU, in MHz.

        Returns:
            Any: A scpufreq object that contains the current frequency of the CPU.
        """
        return psutil.cpu_freq()

    @staticmethod
    def CpuFreqPerCore() -> list:
        """Returns the current frequency per CPU core, in MHz.

        Returns:
            list: A list of scpufreq objects, one per CPU core.
        """
        return psutil.cpu_freq(True)

    @staticmethod
    def CpuLoadAvg() -> list:
        """Returns the 1, 5, and 15 minute load averages for the CPU system.

        Returns:
            list: list of 3 floats representing the load averages over 1, 5, and 15 minutes.
        """
        return psutil.getloadavg()

    @staticmethod
    def CpuAffinity(pid: int = None) -> Union[list, None]:
        """Returns a list of CPU cores the given process (or current process if no PID is specified) is allowed to run on.

        Args:
            pid (int, optional): The process ID. If None, the current process is used. Defaults to None.

        Returns:
            Union[list, None]: A list of CPU cores the process is allowed to use, or None if the process is not found.
        """
        if pid:
            p = psutil.Process(pid)
            return p.cpu_affinity()
        return psutil.Process().cpu_affinity()

    @staticmethod
    def CpuTimesPerCore() -> list:
        """Returns the CPU times per core in the system.

        Returns:
            list: A list of scputimes objects per CPU core.
        """
        return psutil.cpu_times(percpu=True)

    @staticmethod
    def CpuUsagePerCore(interval: Union[None, float] = 1.0) -> list:
        """Returns the CPU usage per core over the given interval.

        Args:
            interval (Union[None, float], optional): Interval for calculating the usage. Defaults to 1.0.

        Returns:
            list: list of CPU usage percentages per core.
        """
        return psutil.cpu_percent(interval=interval, percpu=True)
    
    if LINUX:
        @staticmethod
        def CpuWaitTime() -> float:
            """Returns the CPU I/O wait time percentage.

            Returns:
                float: The percentage of time the CPU spends waiting for I/O to complete.
            """
            return psutil.cpu_times().iowait

        @staticmethod
        def CpuSteal() -> float:
            """Returns the percentage of time the CPU is stolen by other operating systems (in virtualized environments).

            Returns:
                float: The percentage of time the CPU is stolen.
            """
            return psutil.cpu_times().steal

class Battery:
    @staticmethod
    def BatteryStatus() -> str:
        """Returns the status of the battery (e.g., 'charging', 'discharging', 'full', or 'not present').

        Returns:
            str: The battery status.
        """
        battery = psutil.sensors_battery()
        if battery is None:
            return "Battery not present"
        return battery.power_plugged and "Charging" or "Discharging" if battery.percent < 100 else "Full"

    @staticmethod
    def BatteryPercentage() -> float:
        """Returns the current battery percentage.

        Returns:
            float: The battery percentage.
        """
        battery = psutil.sensors_battery()
        if battery is None:
            return -1.0  # Battery not present
        return battery.percent

    @staticmethod
    def BatteryTimeLeft() -> str:
        """Returns the estimated time left on the battery in minutes. If plugged in, returns 'Charging'.

        Returns:
            str: Time left + Time Period (Minutes/Hours/Seconds/Days/Months/Years/Centuries), or 'Charging' if plugged in.
        """
        battery = psutil.sensors_battery()
        if battery is None:
            return "Battery not present"
        if battery.power_plugged:
            return "Charging"
        if battery.secsleft != psutil.POWER_TIME_UNKNOWN and battery.secsleft != 4294967295:
            mins = battery.secsleft / 60
            if mins > 60:
                hours = mins / 60
                if hours > 24:
                    days = hours / 24
                    if days > 31:
                        months = days / 31
                        if months > 12:
                            years = months / 12
                            if years > 100:
                                century = years / 100
                                return str(century) + " Centuries"
                            else:
                                return str(years) + " Years"
                        return str(months) + " Months"
                    else:
                        return str(days) + " Days"
                else:
                    return str(hours) + " Hours"
            else:
                return str(mins) + " Minutes"

        else: return "-1.0" + " Unknown Duration"

    @staticmethod
    def BatteryPlugged() -> bool:
        """Returns True if the battery is currently plugged in (charging), otherwise False.

        Returns:
            bool: True if the battery is plugged in, else False.
        """
        battery = psutil.sensors_battery()
        if battery is None:
            return False  # No battery present
        return battery.power_plugged

    @staticmethod
    def BatterySecsLeft() -> Union[int, None]:
        """Returns the remaining seconds of battery life. Returns None if the time left is unknown.

        Returns:
            Union[int, None]: Remaining time in seconds or None if unknown.
        """
        battery = psutil.sensors_battery()
        if battery is None:
            return None  # No battery present
        return battery.secsleft if battery.secsleft != psutil.POWER_TIME_UNKNOWN and battery.secsleft != 4294967295 else None

    @staticmethod
    def BatteryPluggedTime() -> float:
        """Returns the time in seconds that the battery has been plugged in.

        Returns:
            float: The time in seconds the battery has been plugged in.
        """
        battery = psutil.sensors_battery()
        if battery is None:
            return -1.0  # Battery not present
        try:
            return battery.plugged_time if battery.plugged_time != psutil.POWER_TIME_UNKNOWN else -10.0
        except AttributeError:
            return -11.0

    @staticmethod
    def BatteryIsCharging() -> bool:
        """Returns True if the battery is currently charging, False otherwise.

        Returns:
            bool: True if the battery is charging, False otherwise.
        """
        battery = psutil.sensors_battery()
        if battery is None:
            return False  # No battery present
        return battery.power_plugged

    @staticmethod
    def BatteryTimeToFullCharge() -> Union[str, float]:
        """Returns the estimated time in minutes for the battery to be fully charged. Returns 'Charging' if plugged in.

        Returns:
            Union[str, float]: Time to full charge in minutes or 'Charging' if plugged in.
        """
        battery = psutil.sensors_battery()
        if battery is None:
            return "Battery not present"
        if battery.power_plugged:
            return "Charging"
        return battery.secsleft / 60 if battery.secsleft != psutil.POWER_TIME_UNKNOWN else -1.0

    @staticmethod
    def BatteryDetails() -> Union[str, dict]:
        """Returns detailed information about the battery, including percent, time left, and charging status.

        Returns:
            Union[str, dict]: A dictionary with detailed battery information or 'Battery not present'.
        """
        battery = psutil.sensors_battery()
        if battery is None:
            return "Battery not present"
        return {
            "Percent": battery.percent,
            "Time Left": battery.secsleft / 60 if battery.secsleft != psutil.POWER_TIME_UNKNOWN else "Unknown",
            "Charging": battery.power_plugged
        }

    @staticmethod
    def BatteryStatusDetails() -> Union[str, dict]:
        """Returns detailed battery status information, including charging status and percent.

        Returns:
            Union[str, dict]: A dictionary with battery status details or 'Battery not present'.
        """
        battery = psutil.sensors_battery()
        if battery is None:
            return "Battery not present"
        return {
            "Status": battery.power_plugged and "Charging" or "Discharging",
            "Percent": battery.percent,
            "Time Left": battery.secsleft / 60 if battery.secsleft != psutil.POWER_TIME_UNKNOWN else "Unknown"
        }

    @staticmethod
    def BatteryType() -> str:
        """Returns the battery type (e.g., 'Li-ion', 'NiMH', etc.) if available.

        Returns:
            str: The battery type or 'Unknown' if the information is unavailable.
        """
        # Note: This may not be available on all systems.
        try:
            battery = psutil.sensors_battery()
            if battery is None:
                return "Battery not present"
            # Some systems might provide battery type info in different ways, and this can be platform-dependent
            # In many cases, this info is not directly available through psutil.
            return "Li-ion"  # Placeholder, may need custom detection
        except Exception as e:
            return "Unknown"  # Return 'Unknown' if unable to get battery type

#Linux Only
class Temperature:
        """Linux Only"""
        @staticmethod
        def CpuTemp() -> float:
            """Returns the current temperature of the CPU in Celsius.
            
            Returns:
                float: The current CPU temperature in Celsius.
            """
            sensors = psutil.sensors_temperature()
            return sensors.get('cpu_thermal', [])[0].current if 'cpu_thermal' in sensors else -1.0
        
        @staticmethod
        def GpuTemp() -> float:
            """Returns the current temperature of the GPU in Celsius.
            
            Returns:
                float: The current GPU temperature in Celsius.
            """
            sensors = psutil.sensors_temperatures()
            return sensors.get('gpu', [])[0].current if 'gpu' in sensors else -1.0

        @staticmethod
        def TempSensors() -> dict:
            """Returns a dictionary of available temperature sensors and their values.
            
            Returns:
                dict: A dictionary where keys are sensor names, and values are lists of sensor readings.
            """
            return psutil.sensors_temperatures()

        @staticmethod
        def TempMax() -> float:
            """Returns the maximum temperature of the CPU in Celsius.
            
            Returns:
                float: The maximum CPU temperature.
            """
            sensors = psutil.sensors_temperatures()
            return max([sensor.current for sensor in sensors.get('cpu_thermal', [])]) if 'cpu_thermal' in sensors else -1.0

        @staticmethod
        def TempMin() -> float:
            """Returns the minimum temperature of the CPU in Celsius.
            
            Returns:
                float: The minimum CPU temperature.
            """
            sensors = psutil.sensors_temperatures()
            return min([sensor.current for sensor in sensors.get('cpu_thermal', [])]) if 'cpu_thermal' in sensors else -1.0

        @staticmethod
        def TempAveraged() -> float:
            """Returns the average temperature of the CPU in Celsius.
            
            Returns:
                float: The average CPU temperature.
            """
            sensors = psutil.sensors_temperatures()
            temperatures = [sensor.current for sensor in sensors.get('cpu_thermal', [])] if 'cpu_thermal' in sensors else []
            return sum(temperatures) / len(temperatures) if temperatures else -1.0

        @staticmethod
        def CpuTempPerCore() -> list:
            """Returns the temperature of each CPU core in Celsius.
            
            Returns:
                list: A list of CPU temperatures for each core.
            """
            sensors = psutil.sensors_temperatures()
            return [sensor.current for sensor in sensors.get('cpu_thermal', [])] if 'cpu_thermal' in sensors else []

        @staticmethod
        def GpuTempPerCore() -> list:
            """Returns the temperature of each GPU core in Celsius.
            
            Returns:
                list: A list of GPU temperatures for each core.
            """
            sensors = psutil.sensors_temperatures()
            return [sensor.current for sensor in sensors.get('gpu', [])] if 'gpu' in sensors else []

        @staticmethod
        def TempCritical() -> float:
            """Returns the critical temperature threshold for the CPU in Celsius.
            
            Returns:
                float: The critical temperature threshold.
            """
            sensors = psutil.sensors_temperatures()
            return max([sensor.critical for sensor in sensors.get('cpu_thermal', [])]) if 'cpu_thermal' in sensors else -1.0

        @staticmethod
        def TempWarning() -> float:
            """Returns the warning temperature threshold for the CPU in Celsius.
            
            Returns:
                float: The warning temperature threshold.
            """
            sensors = psutil.sensors_temperatures()
            return max([sensor.warning for sensor in sensors.get('cpu_thermal', [])]) if 'cpu_thermal' in sensors else -1.0

        @staticmethod
        def TempStatus() -> dict:
            """Returns a dictionary of temperature status (e.g., current, critical, warning) for each sensor.
            
            Returns:
                dict: A dictionary with temperature status.
            """
            sensors = psutil.sensors_temperatures()
            return {sensor: {'current': s.current, 'critical': s.critical, 'warning': s.warning} for sensor, s in sensors.items()}

        @staticmethod
        def GpuTempStatus() -> dict:
            """Returns the temperature status for the GPU.
            
            Returns:
                dict: A dictionary with GPU temperature status.
            """
            sensors = psutil.sensors_temperatures()
            return {'gpu': {'current': sensors.get('gpu', [])[0].current, 'critical': sensors.get('gpu', [])[0].critical}} if 'gpu' in sensors else {}

        @staticmethod
        def CpuTempStatus() -> dict:
            """Returns the temperature status for the CPU.
            
            Returns:
                dict: A dictionary with CPU temperature status.
            """
            sensors = psutil.sensors_temperatures()
            return {'cpu': {'current': sensors.get('cpu_thermal', [])[0].current, 'critical': sensors.get('cpu_thermal', [])[0].critical}} if 'cpu_thermal' in sensors else {}

        @staticmethod
        def MaxTempThreshold() -> float:
            """Returns the maximum temperature threshold for all sensors.
            
            Returns:
                float: The maximum temperature threshold across all sensors.
            """
            sensors = psutil.sensors_temperatures()
            all_temps = [sensor.critical for sensor in sensors.get('cpu_thermal', [])] if 'cpu_thermal' in sensors else []
            return max(all_temps) if all_temps else -1.0

        @staticmethod
        def MinTempThreshold() -> float:
            """Returns the minimum temperature threshold for all sensors.
            
            Returns:
                float: The minimum temperature threshold across all sensors.
            """
            sensors = psutil.sensors_temperatures()
            all_temps = [sensor.warning for sensor in sensors.get('cpu_thermal', [])] if 'cpu_thermal' in sensors else []
            return min(all_temps) if all_temps else -1.0

        @staticmethod
        def TemperatureTrends() -> dict:
            """Returns the temperature trends (current, max, min) for each sensor.
            
            Returns:
                dict: A dictionary with temperature trends for each sensor.
            """
            sensors = psutil.sensors_temperatures()
            trends = {}
            for sensor_name, sensor_list in sensors.items():
                trends[sensor_name] = {
                    'current': max([s.current for s in sensor_list]),
                    'max': max([s.max for s in sensor_list]),
                    'min': min([s.min for s in sensor_list]),
                }
            return trends

        @staticmethod
        def TempInFahrenheit() -> dict:
            """Returns all temperature readings in Fahrenheit.
            
            Returns:
                dict: A dictionary with temperature readings in Fahrenheit.
            """
            sensors = psutil.sensors_temperatures()
            return {sensor: {'current': (s.current * 9 / 5) + 32} for sensor, s in sensors.items()}

        @staticmethod
        def TempInKelvin() -> dict:
            """Returns all temperature readings in Kelvin.
            
            Returns:
                dict: A dictionary with temperature readings in Kelvin.
            """
            sensors = psutil.sensors_temperatures()
            return {sensor: {'current': s.current + 273.15} for sensor, s in sensors.items()}

        @staticmethod
        def TempForProcesses(pid: int) -> dict:
            """Returns the temperature readings for processes, if available.
            
            Args:
                pid (int): The process ID.
            
            Returns:
                dict: A dictionary with temperature readings for the given process.
            """
            sensors = psutil.sensors_temperatures()
            return {pid: {'current': sensors.get('cpu_thermal', [])[0].current}} if 'cpu_thermal' in sensors else {}

class Disk:
    @staticmethod
    def DiskUsage(path: str = '/') -> dict:
        """
        Returns the disk usage (free, used, total, percent).
        Works cross-platform (Linux, Windows, macOS)
        """
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.disk_usage(path)._asdict()
            except Exception as e:
                print(f"Error fetching disk usage: {e}")
                return {}
        return {}

    @staticmethod
    def DiskPartitions() -> list:
        """
        Returns a list of partitions on the system.
        Works cross-platform (Linux, Windows, macOS)
        """
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.disk_partitions()
            except Exception as e:
                print(f"Error fetching partitions: {e}")
                return []
        return []

    @staticmethod
    def DiskFree(path: str = '/') -> float:
        """
        Returns the free space of a partition in bytes.
        Works cross-platform (Linux, Windows, macOS)
        """
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.disk_usage(path).free
            except Exception as e:
                print(f"Error fetching free space: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def DiskUsed(path: str = '/') -> float:
        """
        Returns the used space of a partition in bytes.
        Works cross-platform (Linux, Windows, macOS)
        """
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.disk_usage(path).used
            except Exception as e:
                print(f"Error fetching used space: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def DiskTotal(path: str = '/') -> float:
        """
        Returns the total space of a partition in bytes.
        Works cross-platform (Linux, Windows, macOS)
        """
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.disk_usage(path).total
            except Exception as e:
                print(f"Error fetching total space: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def DiskReadBytes() -> int:
        """
        Returns the total number of bytes read from all disks.
        Works cross-platform (Linux, Windows, macOS)
        """
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.disk_io_counters().read_bytes
            except Exception as e:
                print(f"Error fetching read bytes: {e}")
                return 0
        return 0

    @staticmethod
    def DiskWriteBytes() -> int:
        """
        Returns the total number of bytes written to all disks.
        Works cross-platform (Linux, Windows, macOS)
        """
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.disk_io_counters().write_bytes
            except Exception as e:
                print(f"Error fetching write bytes: {e}")
                return 0
        return 0

    @staticmethod
    def DiskReads() -> int:
        """
        Returns the total number of read operations performed on all disks.
        Works cross-platform (Linux, Windows, macOS)
        """
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.disk_io_counters().read_count
            except Exception as e:
                print(f"Error fetching read count: {e}")
                return 0
        return 0

    @staticmethod
    def DiskWrites() -> int:
        """
        Returns the total number of write operations performed on all disks.
        Works cross-platform (Linux, Windows, macOS)
        """
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.disk_io_counters().write_count
            except Exception as e:
                print(f"Error fetching write count: {e}")
                return 0
        return 0

    @staticmethod
    def DiskReadTime() -> int:
        """
        Returns the total time spent reading from the disk (in milliseconds).
        Works cross-platform (Linux, Windows, macOS)
        """
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.disk_io_counters().read_time
            except Exception as e:
                print(f"Error fetching read time: {e}")
                return 0
        return 0

    @staticmethod
    def DiskWriteTime() -> int:
        """
        Returns the total time spent writing to the disk (in milliseconds).
        Works cross-platform (Linux, Windows, macOS)
        """
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.disk_io_counters().write_time
            except Exception as e:
                print(f"Error fetching write time: {e}")
                return 0
        return 0

    @staticmethod
    def DiskIOMerges() -> int:
        """
        Returns the total number of merged I/O operations.
        Works on Linux (Not available on Windows/macOS)
        """
        if SYSTEM == 'linux':
            try:
                return psutil.disk_io_counters().merged_read_count + psutil.disk_io_counters().merged_write_count
            except Exception as e:
                print(f"Error fetching I/O merges: {e}")
                return 0
        return 0

    @staticmethod
    def DiskQueueDepth() -> int:
        """
        Returns the current disk I/O queue depth.
        Works on Linux (Not available on Windows/macOS)
        """
        if SYSTEM == 'linux':
            try:
                return psutil.disk_io_counters().queue
            except Exception as e:
                print(f"Error fetching disk queue depth: {e}")
                return 0
        return 0

    @staticmethod
    def DiskHealth(device: str) -> dict:
        """
        Returns the SMART health status of a disk.
        Works on Linux (using smartmontools), Windows (using OpenHardwareMonitor), macOS (via system commands)
        """
        if SYSTEM == 'linux':
            try:
                result = subprocess.check_output(["smartctl", "-H", device])
                return result.decode()
            except subprocess.CalledProcessError as e:
                print(f"Error fetching SMART data for {device}: {e}")
                return {}

        elif SYSTEM == 'windows':
            try:
                result = subprocess.check_output(["OpenHardwareMonitorCLI", "get", "disk", device])
                return result.decode()
            except subprocess.CalledProcessError as e:
                print(f"Error fetching SMART data for {device}: {e}")
                return {}

        elif SYSTEM == 'darwin':
            try:
                result = subprocess.check_output(["diskutil", "info", device])
                return result.decode()
            except subprocess.CalledProcessError as e:
                print(f"Error fetching SMART data for {device}: {e}")
                return {}

        return {}

    @staticmethod
    def DiskTemp(device: str) -> float:
        """
        Returns the temperature of the disk (if supported).
        Works on Linux (using smartmontools), Windows (using OpenHardwareMonitor), macOS (via system commands)
        """
        if SYSTEM == 'linux':
            try:
                result = subprocess.check_output(["smartctl", "-A", device])
                for line in result.decode().splitlines():
                    if "Temperature_Celsius" in line:
                        return float(line.split()[-1])
            except subprocess.CalledProcessError as e:
                print(f"Error fetching temperature for {device}: {e}")
                return -1.0

        elif SYSTEM == 'windows':
            try:
                result = subprocess.check_output(["OpenHardwareMonitorCLI", "get", "temperature", device])
                return float(result.decode())
            except subprocess.CalledProcessError as e:
                print(f"Error fetching temperature for {device}: {e}")
                return -1.0

        elif SYSTEM == 'darwin':
            try:
                result = subprocess.check_output(["sysctl", "-a"])
                return -1.0  # Modify to extract disk temperature on macOS if needed
            except subprocess.CalledProcessError as e:
                print(f"Error fetching temperature for {device}: {e}")
                return -1.0

        return -1.0

class Memory:
    # RAM Functions

    @staticmethod
    def RAMInfo() -> dict:
        """Returns overall system RAM information (total, available, used, percent)."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.virtual_memory()._asdict()
            except Exception as e:
                print(f"Error fetching RAM info: {e}")
                return {}
        return {}

    @staticmethod
    def RAMTotal() -> float:
        """Returns total RAM in bytes."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.virtual_memory().total
            except Exception as e:
                print(f"Error fetching total RAM: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def RAMAvailable() -> float:
        """Returns available RAM in bytes."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.virtual_memory().available
            except Exception as e:
                print(f"Error fetching available RAM: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def RAMUsed() -> float:
        """Returns used RAM in bytes."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.virtual_memory().used
            except Exception as e:
                print(f"Error fetching used RAM: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def RAMPercent() -> float:
        """Returns percentage of RAM used."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.virtual_memory().percent
            except Exception as e:
                print(f"Error fetching RAM percent: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def RAMActive() -> float:
        """Returns active RAM in bytes (used for active processes)."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.virtual_memory().active
            except Exception as e:
                print(f"Error fetching active RAM: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def RAMBuffered() -> float:
        """Returns buffered RAM in bytes (used for temporary caching)."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.virtual_memory().buffers
            except Exception as e:
                print(f"Error fetching buffered RAM: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def RAMShared() -> float:
        """Returns shared RAM in bytes (used by multiple processes)."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.virtual_memory().shared
            except Exception as e:
                print(f"Error fetching shared RAM: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def RAMSlab() -> float:
        """Returns slab memory in bytes (kernel memory used to cache objects)."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.virtual_memory().slab
            except Exception as e:
                print(f"Error fetching slab memory: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def RAMFree() -> float:
        """Returns free RAM in bytes."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.virtual_memory().free
            except Exception as e:
                print(f"Error fetching free RAM: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def RAMUsedByProcesses() -> list:
        """Returns memory used by processes in a list of dictionaries."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return [{"pid": p.info["pid"], "memory": p.info["memory_info"].rss} for p in psutil.process_iter(['pid', 'memory_info'])]
            except Exception as e:
                print(f"Error fetching used RAM by processes: {e}")
                return []
        return []

    @staticmethod
    def RAMSwapTotal() -> float:
        """Returns total swap memory used for RAM overflow (if applicable)."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.swap_memory().total
            except Exception as e:
                print(f"Error fetching swap total memory: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def RAMSwapUsed() -> float:
        """Returns used swap memory in bytes."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.swap_memory().used
            except Exception as e:
                print(f"Error fetching swap used memory: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def RAMSwapFree() -> float:
        """Returns free swap memory in bytes."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.swap_memory().free
            except Exception as e:
                print(f"Error fetching swap free memory: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def RAMSwapPercent() -> float:
        """Returns swap memory usage percentage."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.swap_memory().percent
            except Exception as e:
                print(f"Error fetching swap percent memory: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def RAMSwapInUse() -> bool:
        """Returns True if swap memory is in use, otherwise False."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.swap_memory().used > 0
            except Exception as e:
                print(f"Error checking if swap is in use: {e}")
                return False
        return False

    @staticmethod
    def RAMBufferInfo() -> dict:
        """Returns detailed memory buffer information."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.virtual_memory()._asdict()
            except Exception as e:
                print(f"Error fetching buffer info: {e}")
                return {}
        return {}

    @staticmethod
    def RAMPhysicalMemory() -> float:
        """Returns total physical memory (RAM) in bytes."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                return psutil.virtual_memory().total
            except Exception as e:
                print(f"Error fetching physical memory: {e}")
                return 0.0
        return 0.0

    @staticmethod
    def RAMActiveProcessMemory(pid: int) -> float:
        """Returns memory used by a specific process (pid)."""
        if SYSTEM in ['linux', 'windows', 'darwin']:
            try:
                process = psutil.Process(pid)
                return process.memory_info().rss
            except Exception as e:
                print(f"Error fetching memory for process {pid}: {e}")
                return 0.0
        return 0.0

#Linux Only
class Fan:
        """Linux Only"""
        @staticmethod
        def FansInfo() -> dict:
            """Returns detailed information about all fans, including fan ID and speed (RPM)."""
            fans = psutil.sensors_fans()
            if fans:
                return fans
            return {}

        @staticmethod
        def FansCount() -> int:
            """Returns the number of fans in the system."""
            fans = psutil.sensors_fans()
            if fans:
                return len(fans)
            return 0

        @staticmethod
        def FanSpeed(fan_id: int) -> Union[int, None]:
            """Returns the speed of a specific fan by fan ID (RPM)."""
            fans = psutil.sensors_fans()
            if fans and fan_id in fans:
                return fans[fan_id][0]  # Return the speed (RPM) of the specified fan
            return None

        @staticmethod
        def AllFanSpeeds() -> dict:
            """Returns a dictionary of fan speeds (RPM) for all fans."""
            fans = psutil.sensors_fans()
            if fans:
                return {fan_id: fan[0] for fan_id, fan in fans.items()}
            return {}

        @staticmethod
        def FanInfoById(fan_id: int) -> Union[dict, None]:
            """Returns detailed information about a specific fan (by fan ID)."""
            fans = psutil.sensors_fans()
            if fans and fan_id in fans:
                return {"fan_id": fan_id, "speed_rpm": fans[fan_id][0]}  # Fan ID and speed in RPM
            return None

class Network:
    # 1. Get network interfaces (Linux, Windows)
    @staticmethod
    def Interfaces() -> Union[dict[str, dict], str]:
        """Returns a dictionary of network interfaces, their addresses, and stats. (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            return psutil.net_if_addrs()
        except Exception as e:
            return f"Error: Unable to fetch interfaces - {str(e)}"

    # 2. Get network interface stats (Linux, Windows)
    @staticmethod
    def InterfaceStats() -> Union[dict[str, dict], str]:
        """Returns the stats (bytes, packets, errors, drops, etc.) for each network interface. (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            return psutil.net_if_stats()
        except Exception as e:
            return f"Error: Unable to fetch interface stats - {str(e)}"

    # 3. Get network connections (Linux, Windows)
    @staticmethod
    def NetworkConnections(kind: str = 'inet') -> Union[list, str]:
        """Returns a list of network connections of the specified type (TCP, UDP, etc.). (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            return psutil.net_connections(kind=kind)
        except Exception as e:
            return f"Error: Unable to fetch network connections - {str(e)}"

    # 4. Get network statistics (Linux, Windows)
    @staticmethod
    def NetworkStats() -> Union[dict[str, Union[int, float]], str]:
        """Returns global network statistics like bytes sent, received, errors, etc. (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            return psutil.net_io_counters(pernic=False)._asdict()
        except Exception as e:
            return f"Error: Unable to fetch network stats - {str(e)}"

    # 5. Get per-interface network statistics (Linux, Windows)
    @staticmethod
    def InterfaceNetworkStats(interface: str) -> Union[dict, str]:
        """Returns per-interface network stats (bytes sent, received, etc.). (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            return psutil.net_io_counters(pernic=True).get(interface, {})
        except Exception as e:
            return f"Error: Unable to fetch network stats for {interface} - {str(e)}"

    # 6. Get the default gateway (Linux, Windows)
    @staticmethod
    def DefaultGateway() -> Union[dict[str, str], str]:
        """Returns the default gateway for the system. (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            gateways = psutil.net_if_addrs()
            return gateways.get('default', {})
        except Exception as e:
            return f"Error: Unable to fetch default gateway - {str(e)}"

    # 7. Get the DNS configuration (Linux, Windows)
    @staticmethod
    def DNSConfig() -> Union[dict[str, list], str]:
        """Returns DNS configuration for the system (DNS servers). (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            dns_info = psutil.net_if_addrs()
            return dns_info.get('dns', {})
        except Exception as e:
            return f"Error: Unable to fetch DNS configuration - {str(e)}"

    # 8. Get the IP address for a given interface (Linux, Windows)
    @staticmethod
    def IPAddress(interface: str) -> Union[str, str]:
        """Returns the IP address of a given interface. (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            addrs = psutil.net_if_addrs().get(interface, [])
            for addr in addrs:
                if addr.family == socket.AF_INET:
                    return addr.address
            return "Error: No IPv4 address found for this interface."
        except Exception as e:
            return f"Error: Unable to fetch IP address for {interface} - {str(e)}"

    # 9. Get MAC address for a given interface (Linux, Windows)
    @staticmethod
    def MACAddress(interface: str) -> Union[str, str]:
        """Returns the MAC address of a given interface. (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            addrs = psutil.net_if_addrs().get(interface, [])
            for addr in addrs:
                if addr.family == psutil.AF_LINK:
                    return addr.address
            return "Error: No MAC address found for this interface."
        except Exception as e:
            return f"Error: Unable to fetch MAC address for {interface} - {str(e)}"

    # 10. Get the hostname of the machine (Linux, Windows)
    @staticmethod
    def Hostname() -> Union[str, str]:
        """Returns the hostname of the machine. (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            return socket.gethostname()
        except Exception as e:
            return f"Error: Unable to fetch hostname - {str(e)}"

    # 11. Get the fully qualified domain name (FQDN) (Linux, Windows)
    @staticmethod
    def FQDN() -> Union[str, str]:
        """Returns the fully qualified domain name (FQDN) of the machine. (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            return socket.getfqdn()
        except Exception as e:
            return f"Error: Unable to fetch FQDN - {str(e)}"

    # 12. Get the IP address of the current machine (Linux, Windows)
    @staticmethod
    def LocalIPAddress() -> Union[str, str]:
        """Returns the local IP address of the machine. (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            hostname = socket.gethostname()
            return socket.gethostbyname(hostname)
        except Exception as e:
            return f"Error: Unable to fetch local IP address - {str(e)}"

    # 13. Get current network interface state (Linux, Windows)
    @staticmethod
    def InterfaceState(interface: str) -> Union[str, str]:
        """Returns the current state (UP/DOWN) of the network interface. (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            stats = psutil.net_if_stats().get(interface)
            if stats:
                return "UP" if stats.isup else "DOWN"
            return f"Error: Interface {interface} not found."
        except Exception as e:
            return f"Error: Unable to fetch interface state for {interface} - {str(e)}"

    # 14. Check if a network interface is up (Linux, Windows)
    @staticmethod
    def IsInterfaceUp(interface: str) -> Union[bool, str]:
        """Returns True if the network interface is up, False otherwise. (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            stats = psutil.net_if_stats().get(interface)
            return stats.isup if stats else f"Error: Interface {interface} not found."
        except Exception as e:
            return f"Error: Unable to check if interface {interface} is up - {str(e)}"

    # 15. Get network connection by PID (Linux, Windows)
    @staticmethod
    def NetworkConnectionsByPID(pid: int) -> Union[list, str]:
        """Returns network connections by the process ID (PID). (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            return psutil.net_connections(kind='inet', pid=pid)
        except Exception as e:
            return f"Error: Unable to fetch network connections for PID {pid} - {str(e)}"

    # 16. Get the local ports in use (Linux, Windows)
    @staticmethod
    def LocalPortsInUse() -> Union[list, str]:
        """Returns a list of local ports in use. (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            connections = psutil.net_connections(kind='inet')
            return [conn.laddr.port for conn in connections]
        except Exception as e:
            return f"Error: Unable to fetch local ports - {str(e)}"

    # 17. Get the external IP address (Linux, Windows)
    @staticmethod
    def ExternalIPAddress() -> Union[str, str]:
        """Returns the external IP address of the system (if available)."""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            import requests
            response = requests.get("https://api.ipify.org")
            return response.text
        except Exception as e:
            return f"Error: Unable to fetch external IP address - {str(e)}"

    # 18. Get network interface type (Linux, Windows)
    @staticmethod
    def InterfaceType(interface: str) -> Union[str, str]:
        """Returns the type of a network interface (e.g., ethernet, wifi). (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            addrs = psutil.net_if_addrs().get(interface, [])
            if addrs:
                return addrs[0].family.name
            return f"Error: Interface {interface} not found."
        except Exception as e:
            return f"Error: Unable to fetch interface type for {interface} - {str(e)}"

    # 19. Get network connection status (Linux, Windows)
    @staticmethod
    def ConnectionStatus(connection) -> Union[str, str]:
        """Returns the status of a network connection (ESTABLISHED, LISTEN, etc.). (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            return connection.status
        except Exception as e:
            return f"Error: Unable to fetch connection status - {str(e)}"

    # 20. Get the netmask of an interface (Linux, Windows)
    @staticmethod
    def Netmask(interface: str) -> Union[str, str]:
        """Returns the netmask of the given network interface. (Linux, Windows only)"""
        if SYSTEM not in ['linux', 'windows']:
            return "Error: This function is only supported on Linux and Windows."
        try:
            addrs = psutil.net_if_addrs().get(interface, [])
            for addr in addrs:
                if addr.family == socket.AF_INET:
                    return addr.netmask
            return f"Error: No netmask found for {interface}."
        except Exception as e:
            return f"Error: Unable to fetch netmask for {interface} - {str(e)}"

class System(CPU, Battery, Temperature, Disk, Memory, Fan, Network):
    #CPU
    @classmethod
    def CpuUsage(cls, interval: Union[None, float] = 1.0) -> float:
        """This function calculates the CPU usage via the given interval and returns the CPU usage.

        Args:
            interval (Union[None, float], optional): Interval defines time used to calculate the CPU usage. None defines the time from module import. Defaults to 1.0.

        Returns:
            float: The Cpu usage.
        """
        return super().CpuUsage(interval)

    @classmethod
    def CpuUsageDetails(cls):
        """This function returns the following details about the CPU usage ->

        user: This represents the time spent by normal processes executing in the user mode.
        system: This represents the time spent by processes executing in the kernel mode.
        idle: This represents the time when the system is idle.
        nice: This represents the time spent by priority processes executing in the user mode.
        iowait: This represents the time spent waiting for I/O to complete.
        irq: This represents the time spent for servicing hardware interrupts.
        softirq: This represents the time spent for servicing software interrupts.
        steal: Represents the time spent by other operating systems running in a virtualized environment
        guest: This represents the time spent running a virtual CPU for guest operating systems under the control of the Linux kernel.

        WINDOWS ONLY:
        interrupt: This represents the time spent for servicing hardware interrupts.
        dpc: This represents the time spent servicing deferred procedure calls (DPCs).

        Returns:
            Any: Return type -> psutil._ps[Your OS].scputimes
        """
        return super().CpuUsageDetails()

    @classmethod
    def CpuUsageTimesDetails(cls, interval: Union[float, None] = 1.0):
        """Same as CpuUsage but returns like CpuUsageDetails.

        Args:
            interval (Union[None, float], optional): Interval defines time used to calculate the CPU usage. None defines the time from module import. Defaults to 1.0.

        Returns:
            Any: Return Type -> psutil._ps[Your OS].scputimes.
        """
        return super().CpuUsageTimesDetails(interval)

    @classmethod
    def CpuCount(cls, logical: bool = False) -> int:
        """Returns the number of Cpu Cores in System.

        Args:
            logical (bool, optional): If set to true returned value will be [Number of Physical Cores + Hyper Threading]. Defaults to False.

        Returns:
            int: The number of Cpu Physical Cores [+ Hyper Threading if [logical] is True].
        """
        return super().CpuCount(logical)

    @classmethod
    def CpuStats(cls):
        """Returns detailed CPU statistics such as context switches, interrupts, and soft interrupts.

        Returns:
            Any: Object containing detailed statistics about the CPU.
        """
        return super().CpuStats()

    @classmethod
    def CpuFreq(cls):
        """Returns the current frequency of the CPU, in MHz.

        Returns:
            Any: A scpufreq object that contains the current frequency of the CPU.
        """
        return super().CpuFreq()

    @classmethod
    def CpuFreqPerCore(cls) -> list:
        """Returns the current frequency per CPU core, in MHz.

        Returns:
            list: A list of scpufreq objects, one per CPU core.
        """
        return super().CpuFreqPerCore()

    @classmethod
    def CpuLoadAvg(cls) -> list:
        """Returns the 1, 5, and 15 minute load averages for the CPU system.

        Returns:
            list: list of 3 floats representing the load averages over 1, 5, and 15 minutes.
        """
        return super().CpuLoadAvg()

    @classmethod
    def CpuAffinity(cls, pid: int = None) -> Union[list, None]:
        """Returns a list of CPU cores the given process (or current process if no PID is specified) is allowed to run on.

        Args:
            pid (int, optional): The process ID. If None, the current process is used. Defaults to None.

        Returns:
            Union[list, None]: A list of CPU cores the process is allowed to use, or None if the process is not found.
        """
        return super().CpuAffinity(pid)

    @classmethod
    def CpuTimesPerCore(cls) -> list:
        """Returns the CPU times per core in the system.

        Returns:
            list: A list of scputimes objects per CPU core.
        """
        return super().CpuTimesPerCore()

    @classmethod
    def CpuUsagePerCore(cls, interval: Union[None, float] = 1.0) -> list:
        """Returns the CPU usage per core over the given interval.

        Args:
            interval (Union[None, float], optional): Interval for calculating the usage. Defaults to 1.0.

        Returns:
            list: list of CPU usage percentages per core.
        """
        return super().CpuUsagePerCore(interval)

    if LINUX:
        @classmethod
        def CpuWaitTime(cls) -> float:
            """Returns the CPU I/O wait time percentage.

            Returns:
                float: The percentage of time the CPU spends waiting for I/O to complete.
            """
            return super().CpuWaitTime()

        @classmethod
        def CpuSteal(cls) -> float:
            """Returns the percentage of time the CPU is stolen by other operating systems (in virtualized environments).

            Returns:
                float: The percentage of time the CPU is stolen.
            """
            return super().CpuSteal()
    
    #Battery
    @classmethod
    def SystemBatteryStatus(cls) -> str:
        """Returns the battery status using the Battery class."""
        return super().BatteryStatus()

    @classmethod
    def SystemBatteryPercentage(cls) -> float:
        """Returns the battery percentage using the Battery class."""
        return super().BatteryPercentage()

    @classmethod
    def SystemBatteryTimeLeft(cls) -> str:
        """Returns the estimated time left on the battery using the Battery class.
        
        Returns:
            str: Time left + Time Period (Minutes/Hours/Seconds/Days/Months/Years/Centuries), or 'Charging' if plugged in."""
        return super().BatteryTimeLeft()

    @classmethod
    def SystemBatteryPlugged(cls) -> bool:
        """Returns whether the battery is plugged in (charging) using the Battery class."""
        return super().BatteryPlugged()

    @classmethod
    def SystemBatterySecsLeft(cls) -> Union[int, None]:
        """Returns the remaining battery life in seconds using the Battery class."""
        return super().BatterySecsLeft()

    @classmethod
    def SystemBatteryPluggedTime(cls) -> float:
        """Returns the time the battery has been plugged in using the Battery class."""
        return super().BatteryPluggedTime()

    @classmethod
    def SystemBatteryIsCharging(cls) -> bool:
        """Returns whether the battery is charging using the Battery class."""
        return super().BatteryIsCharging()

    @classmethod
    def SystemBatteryTimeToFullCharge(cls) -> Union[str, float]:
        """Returns the estimated time to full charge using the Battery class."""
        return super().BatteryTimeToFullCharge()

    @classmethod
    def SystemBatteryDetails(cls) -> Union[str, dict]:
        """Returns detailed battery information using the Battery class."""
        return super().BatteryDetails()

    @classmethod
    def SystemBatteryStatusDetails(cls) -> Union[str, dict]:
        """Returns detailed battery status information using the Battery class."""
        return super().BatteryStatusDetails()

    @classmethod
    def SystemBatteryType(cls) -> str:
        """Returns the battery type using the Battery class."""
        return super().BatteryType()
    
    #Temperature
    if LINUX:
        @classmethod
        def CpuTemp(cls) -> float:
            """Returns the current temperature of the CPU in Celsius."""
            return super().CpuTemp()

        @classmethod
        def GpuTemp(cls) -> float:
            """Returns the current temperature of the GPU in Celsius."""
            return super().GpuTemp()

        @classmethod
        def TempSensors(cls) -> dict:
            """Returns a dictionary of available temperature sensors and their values."""
            return super().TempSensors()

        @classmethod
        def TempMax(cls) -> float:
            """Returns the maximum temperature of the CPU in Celsius."""
            return super().TempMax()

        @classmethod
        def TempMin(cls) -> float:
            """Returns the minimum temperature of the CPU in Celsius."""
            return super().TempMin()

        @classmethod
        def TempAveraged(cls) -> float:
            """Returns the average temperature of the CPU in Celsius."""
            return super().TempAveraged()

        @classmethod
        def CpuTempPerCore(cls) -> list:
            """Returns the temperature of each CPU core in Celsius."""
            return super().CpuTempPerCore()

        @classmethod
        def GpuTempPerCore(cls) -> list:
            """Returns the temperature of each GPU core in Celsius."""
            return super().GpuTempPerCore()

        @classmethod
        def TempCritical(cls) -> float:
            """Returns the critical temperature threshold for the CPU in Celsius."""
            return super().TempCritical()

        @classmethod
        def TempWarning(cls) -> float:
            """Returns the warning temperature threshold for the CPU in Celsius."""
            return super().TempWarning()

        @classmethod
        def TempStatus(cls) -> dict:
            """Returns a dictionary of temperature status for each sensor."""
            return super().TempStatus()

        @classmethod
        def GpuTempStatus(cls) -> dict:
            """Returns the temperature status for the GPU."""
            return super().GpuTempStatus()

        @classmethod
        def CpuTempStatus(cls) -> dict:
            """Returns the temperature status for the CPU."""
            return super().CpuTempStatus()

        @classmethod
        def MaxTempThreshold(cls) -> float:
            """Returns the maximum temperature threshold for all sensors."""
            return super().MaxTempThreshold()

        @classmethod
        def MinTempThreshold(cls) -> float:
            """Returns the minimum temperature threshold for all sensors."""
            return super().MinTempThreshold()

        @classmethod
        def TemperatureTrends(cls) -> dict:
            """Returns the temperature trends (current, max, min) for each sensor."""
            return super().TemperatureTrends()

        @classmethod
        def TempInFahrenheit(cls) -> dict:
            """Returns all temperature readings in Fahrenheit."""
            return super().TempInFahrenheit()

        @classmethod
        def TempInKelvin(cls) -> dict:
            """Returns all temperature readings in Kelvin."""
            return super().TempInKelvin()

        @classmethod
        def TempForProcesses(cls, pid: int) -> dict:
            """Returns the temperature readings for processes, if available."""
            return super().TempForProcesses(pid)
    
    #Disk
    @classmethod
    def DiskUsage(cls, path: str = '/') -> dict:
        """
        Returns the disk usage (free, used, total, percent).
        Delegates to the Disk class methods.
        """
        return super().DiskUsage(path)

    @classmethod
    def DiskPartitions(cls) -> list:
        """
        Returns a list of partitions on the system.
        Delegates to the Disk class methods.
        """
        return super().DiskPartitions()

    @classmethod
    def DiskFree(cls, path: str = '/') -> float:
        """
        Returns the free space of a partition in bytes.
        Delegates to the Disk class methods.
        """
        return super().DiskFree(path)

    @classmethod
    def DiskUsed(cls, path: str = '/') -> float:
        """
        Returns the used space of a partition in bytes.
        Delegates to the Disk class methods.
        """
        return super().DiskUsed(path)

    @classmethod
    def DiskTotal(cls, path: str = '/') -> float:
        """
        Returns the total space of a partition in bytes.
        Delegates to the Disk class methods.
        """
        return super().DiskTotal(path)

    @classmethod
    def DiskReadBytes(cls) -> int:
        """
        Returns the total number of bytes read from all disks.
        Delegates to the Disk class methods.
        """
        return super().DiskReadBytes()

    @classmethod
    def DiskWriteBytes(cls) -> int:
        """
        Returns the total number of bytes written to all disks.
        Delegates to the Disk class methods.
        """
        return super().DiskWriteBytes()

    @classmethod
    def DiskReads(cls) -> int:
        """
        Returns the total number of read operations performed on all disks.
        Delegates to the Disk class methods.
        """
        return super().DiskReads()

    @classmethod
    def DiskWrites(cls) -> int:
        """
        Returns the total number of write operations performed on all disks.
        Delegates to the Disk class methods.
        """
        return super().DiskWrites()

    @classmethod
    def DiskReadTime(cls) -> int:
        """
        Returns the total time spent reading from the disk (in milliseconds).
        Delegates to the Disk class methods.
        """
        return super().DiskReadTime()

    @classmethod
    def DiskWriteTime(cls) -> int:
        """
        Returns the total time spent writing to the disk (in milliseconds).
        Delegates to the Disk class methods.
        """
        return super().DiskWriteTime()

    @classmethod
    def DiskIOMerges(cls) -> int:
        """
        Returns the total number of merged I/O operations.
        Delegates to the Disk class methods (Linux only).
        """
        return super().DiskIOMerges()

    @classmethod
    def DiskQueueDepth(cls) -> int:
        """
        Returns the current disk I/O queue depth.
        Delegates to the Disk class methods (Linux only).
        """
        return super().DiskQueueDepth()

    @classmethod
    def DiskHealth(cls, device: str) -> dict:
        """
        Returns the SMART health status of a disk.
        Delegates to the Disk class methods.
        """
        return super().DiskHealth(device)

    @classmethod
    def DiskTemp(cls, device: str) -> float:
        """
        Returns the temperature of the disk (if supported).
        Delegates to the Disk class methods.
        """
        return super().DiskTemp(device)
    
    #Memory
    @classmethod
    def RAMInfo(cls) -> dict:
        """Returns overall system RAM information (total, available, used, percent)."""
        return super().RAMInfo()

    @classmethod
    def RAMTotal(cls) -> float:
        """Returns total RAM in bytes."""
        return super().RAMTotal()

    @classmethod
    def RAMAvailable(cls) -> float:
        """Returns available RAM in bytes."""
        return super().RAMAvailable()

    @classmethod
    def RAMUsed(cls) -> float:
        """Returns used RAM in bytes."""
        return super().RAMUsed()

    @classmethod
    def RAMPercent(cls) -> float:
        """Returns percentage of RAM used."""
        return super().RAMPercent()

    @classmethod
    def RAMActive(cls) -> float:
        """Returns active RAM in bytes (used for active processes)."""
        return super().RAMActive()

    @classmethod
    def RAMBuffered(cls) -> float:
        """Returns buffered RAM in bytes (used for temporary caching)."""
        return super().RAMBuffered()

    @classmethod
    def RAMShared(cls) -> float:
        """Returns shared RAM in bytes (used by multiple processes)."""
        return super().RAMShared()

    @classmethod
    def RAMSlab(cls) -> float:
        """Returns slab memory in bytes (kernel memory used to cache objects)."""
        return super().RAMSlab()

    @classmethod
    def RAMFree(cls) -> float:
        """Returns free RAM in bytes."""
        return super().RAMFree()

    @classmethod
    def RAMUsedByProcesses(cls) -> list:
        """Returns memory used by processes in a list of dictionaries."""
        return super().RAMUsedByProcesses()

    @classmethod
    def RAMSwapTotal(cls) -> float:
        """Returns total swap memory used for RAM overflow (if applicable)."""
        return super().RAMSwapTotal()

    @classmethod
    def RAMSwapUsed(cls) -> float:
        """Returns used swap memory in bytes."""
        return super().RAMSwapUsed()

    @classmethod
    def RAMSwapFree(cls) -> float:
        """Returns free swap memory in bytes."""
        return super().RAMSwapFree()

    @classmethod
    def RAMSwapPercent(cls) -> float:
        """Returns swap memory usage percentage."""
        return super().RAMSwapPercent()

    @classmethod
    def RAMSwapInUse(cls) -> bool:
        """Returns True if swap memory is in use, otherwise False."""
        return super().RAMSwapInUse()

    @classmethod
    def RAMBufferInfo(cls) -> dict:
        """Returns detailed memory buffer information."""
        return super().RAMBufferInfo()

    @classmethod
    def RAMPhysicalMemory(cls) -> float:
        """Returns total physical memory (RAM) in bytes."""
        return super().RAMPhysicalMemory()

    @classmethod
    def RAMActiveProcessMemory(cls, pid: int) -> float:
        """Returns memory used by a specific process (pid)."""
        return super().RAMActiveProcessMemory(pid)
    
    #Fans
    if LINUX:
        @classmethod
        def SystemFansInfo(cls) -> dict:
            """Returns detailed information about all fans, including fan ID and speed (RPM)."""
            return super().FansInfo()

        @classmethod
        def SystemFansCount(cls) -> int:
            """Returns the number of fans in the system."""
            return super().FansCount()

        @classmethod
        def SystemFanSpeed(cls, fan_id: int) -> Union[int, None]:
            """Returns the speed of a specific fan by fan ID (RPM)."""
            return super().FanSpeed(fan_id)

        @classmethod
        def SystemAllFanSpeeds(cls) -> dict:
            """Returns a dictionary of fan speeds (RPM) for all fans."""
            return super().AllFanSpeeds()

        @classmethod
        def SystemFanInfoById(cls, fan_id: int) -> Union[dict, None]:
            """Returns detailed information about a specific fan (by fan ID)."""
            return super().FanInfoById(fan_id)
    
    #Network
    
    # 1. Get network interfaces (Linux, Windows)
    @classmethod
    def Interfaces(cls) -> Union[dict[str, dict], str]:
        """Returns a dictionary of network interfaces, their addresses, and stats. (Linux, Windows only)"""
        return super().Interfaces()

    # 2. Get network interface stats (Linux, Windows)
    @classmethod
    def InterfaceStats(cls) -> Union[dict[str, dict], str]:
        """Returns the stats (bytes, packets, errors, drops, etc.) for each network interface. (Linux, Windows only)"""
        return super().InterfaceStats()

    # 3. Get network connections (Linux, Windows)
    @classmethod
    def NetworkConnections(cls, kind: str = 'inet') -> Union[list, str]:
        """Returns a list of network connections of the specified type (TCP, UDP, etc.). (Linux, Windows only)"""
        return super().NetworkConnections(kind)

    # 4. Get network statistics (Linux, Windows)
    @classmethod
    def NetworkStats(cls) -> Union[dict[str, Union[int, float]], str]:
        """Returns global network statistics like bytes sent, received, errors, etc. (Linux, Windows only)"""
        return super().NetworkStats()

    # 5. Get per-interface network statistics (Linux, Windows)
    @classmethod
    def InterfaceNetworkStats(cls, interface: str) -> Union[dict, str]:
        """Returns per-interface network stats (bytes sent, received, etc.). (Linux, Windows only)"""
        return super().InterfaceNetworkStats(interface)

    # 6. Get the default gateway (Linux, Windows)
    @classmethod
    def DefaultGateway(cls) -> Union[dict[str, str], str]:
        """Returns the default gateway for the system. (Linux, Windows only)"""
        return super().DefaultGateway()

    # 7. Get the DNS configuration (Linux, Windows)
    @classmethod
    def DNSConfig(cls) -> Union[dict[str, list], str]:
        """Returns DNS configuration for the system (DNS servers). (Linux, Windows only)"""
        return super().DNSConfig()

    # 8. Get the IP address for a given interface (Linux, Windows)
    @classmethod
    def IPAddress(cls, interface: str) -> Union[str, str]:
        """Returns the IP address of a given interface. (Linux, Windows only)"""
        return super().IPAddress(interface)

    # 9. Get MAC address for a given interface (Linux, Windows)
    @classmethod
    def MACAddress(cls, interface: str) -> Union[str, str]:
        """Returns the MAC address of a given interface. (Linux, Windows only)"""
        return super().MACAddress(interface)

    # 10. Get the hostname of the machine (Linux, Windows)
    @classmethod
    def Hostname(cls) -> Union[str, str]:
        """Returns the hostname of the machine. (Linux, Windows only)"""
        return super().Hostname()

    # 11. Get the fully qualified domain name (FQDN) (Linux, Windows)
    @classmethod
    def FQDN(cls) -> Union[str, str]:
        """Returns the fully qualified domain name (FQDN) of the machine. (Linux, Windows only)"""
        return super().FQDN()

    # 12. Get the IP address of the current machine (Linux, Windows)
    @classmethod
    def LocalIPAddress(cls) -> Union[str, str]:
        """Returns the local IP address of the machine. (Linux, Windows only)"""
        return super().LocalIPAddress()

    # 13. Get current network interface state (Linux, Windows)
    @classmethod
    def InterfaceState(cls, interface: str) -> Union[str, str]:
        """Returns the current state (UP/DOWN) of the network interface. (Linux, Windows only)"""
        return super().InterfaceState(interface)

    # 14. Check if a network interface is up (Linux, Windows)
    @classmethod
    def IsInterfaceUp(cls, interface: str) -> Union[bool, str]:
        """Returns True if the network interface is up, False otherwise. (Linux, Windows only)"""
        return super().IsInterfaceUp(interface)

    # 15. Get network connection by PID (Linux, Windows)
    @classmethod
    def NetworkConnectionsByPID(cls, pid: int) -> Union[list, str]:
        """Returns network connections by the process ID (PID). (Linux, Windows only)"""
        return super().NetworkConnectionsByPID(pid)

    # 16. Get the local ports in use (Linux, Windows)
    @classmethod
    def LocalPortsInUse(cls) -> Union[list, str]:
        """Returns a list of local ports in use. (Linux, Windows only)"""
        return super().LocalPortsInUse()

    # 17. Get the external IP address (Linux, Windows)
    @classmethod
    def ExternalIPAddress(cls) -> Union[str, str]:
        """Returns the external IP address of the system (if available)."""
        return super().ExternalIPAddress()

    # 18. Get network interface type (Linux, Windows)
    @classmethod
    def InterfaceType(cls, interface: str) -> Union[str, str]:
        """Returns the type of a network interface (e.g., ethernet, wifi). (Linux, Windows only)"""
        return super().InterfaceType(interface)

    # 19. Get network connection status (Linux, Windows)
    @classmethod
    def ConnectionStatus(cls, connection) -> Union[str, str]:
        """Returns the status of a network connection (ESTABLISHED, LISTEN, etc.). (Linux, Windows only)"""
        return super().ConnectionStatus(connection)

    # 20. Get the netmask of an interface (Linux, Windows)
    @classmethod
    def Netmask(cls, interface: str) -> Union[str, str]:
        """Returns the netmask of the given network interface. (Linux, Windows only)"""
        return super().Netmask(interface)


import os
import json

LangSetPath = os.path.join(os.path.dirname(os.path.abspath(__file__)), "DEF_LanguageSET.json")
DEF_LANG = os.path.join(os.path.dirname(os.path.abspath(__file__)), "DEF_lang.json")

AVAIL_types = ["elf", "exe", "aef", "caef"]
AVAIL_modes = ["x64", "x86"]

def CompileLangSET() -> dict:
    data  = {}

    with open(LangSetPath, "r") as f:
        data = json.load(f)
        f.close()

    return data

def Compile_LANG(CompiledLangSET: dict, langPath: str=DEF_LANG) -> tuple[dict, dict, list]:
    data:dict = {}

    if langPath == "":
        langPath = DEF_LANG

    with open(langPath, "r") as f:
        data = json.load(f)

    mode = "elf|x86"

    for i, v in enumerate(data.keys()):
        v = v.lower()

        if v in AVAIL_types:
            mode = mode.replace("elf", v)
        
        if v in AVAIL_modes:
            mode = mode.replace("x86", v)

    mode = mode.split("|")

    output_lang = data
    output_langSET = CompiledLangSET[mode[0]][mode[1]]

    return output_lang, output_langSET, mode


# ----------------------------------------------------------------
# Code to ASM/MACHINECODE                                        |
# ----------------------------------------------------------------
# Pheonix Studios                                                |
# ----------------------------------------------------------------
# Akshobhya                                                      |
# ----------------------------------------------------------------
# https://github.com/AkshuDev                                    |
# ----------------------------------------------------------------
import os
import sys

from . import *

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)

from phardwareitk.PLTEC import ASM
from phardwareitk.PLTEC import Reader
from phardwareitk.PLTEC import OBJECT
from phardwareitk.PLTEC import Linker
from phardwareitk.PLTEC import Logger
from phardwareitk.PLTEC import Checker

Checker.PLTEC_initCheck()

def job():
    path_for_file = input("PATH: ")
    path_for_output_file = input("OUTPUT (Don't include extention): ")
    path_for_lang = input("PATH FOR LANG.json (Leave for default): ")
    extr_commands = input("EXTRA COMMANDS (Seperate via ','): ")
    extr_commands = extr_commands.lower().split(",")
    data = ""

    object_ = False

    debug = False

    if "-debug" in extr_commands:
        debug = True

    if '-object' in extr_commands or '-o' in extr_commands:
        object_ = True

    CompiledLangSET = Reader.CompileLangSET()
    Lang, LangSET, mode = Reader.Compile_LANG(CompiledLangSET, path_for_lang)

    if debug:
        Logger.LOG(TitleMsg="Compiled LangSET", errorTitle="LOG", debugMsg=str(LangSET), Message="This is the compiled language set.", descriptionMsg="A language set is a json file including assembly code as the values and the keys as the function names. These json files help to easily allow users to convert custom language by calling these functions."
                , path="PLTEC/main.py", pathMsg="This is the main file of PLTEC Converter", ErrorCode="NoError_This_is_a_log1").log()
        Logger.LOG(TitleMsg="Compiled Lang", errorTitle="LOG", debugMsg=str(Lang), Message="This is the compiled language.", descriptionMsg="A lang/language file is a json file containing all the necessary language syntax to allow the program to convert the custom language to assembly, object, etc. The keys in this file are the syntax of the words/lines and the values are the corrosponding function they call in the language SET."
                , path="PLTEC/main.py", pathMsg="This is the main file of PLTEC Converter", ErrorCode="NoError_This_is_a_log2").log()
        Logger.LOG(TitleMsg="MODE", errorTitle="LOG", debugMsg=str(mode), Message="This is the mode of the program.", descriptionMsg="This is the mode in which the program runs. This mode decides the conversion type of the program."
                , path="PLTEC/main.py", pathMsg="This is the main file of PLTEC Converter", ErrorCode="NoError_This_is_a_log3").log()

    with open(path_for_file, "r") as f:
        data = f.read()
        f.close()

    # Covert data
    Output_data = ""
    Data_SECTION = {}
    cancelled_Sections = 0

    if mode[0] == "elf" and mode[1] == "x86":
        Output_data, Data_SECTION, cancelled_Sections = ASM.x86_Linux.convert_to_asm(data, LangSET, Lang)

    if debug:
        Logger.LOG(TitleMsg="Data Section", errorTitle="LOG", debugMsg=str(Data_SECTION), Message="This is the data section of the assembly output.", descriptionMsg="The data section in assembly refers to a sepecifc section/segment of the program where all the variables are defined."
                , path="PLTEC/main.py", pathMsg="This is the main file of PLTEC Converter", ErrorCode="NoError_This_is_a_log4").log()
        
        Logger.LOG(TitleMsg="Assembly Output", errorTitle="LOG", debugMsg=str(Output_data), Message="This is the assembly output.", descriptionMsg="The assembly file is a specific low level hardware interactive file that contains human readable form of machine code which can be used to do anything with a computer. Assembly is one of the most complex, non-effcient (because of manual memory management), and flexible language in the world. All the other languages are made via assembly. It allows the user to do whatever they want with the computer."
                , path="PLTEC/main.py", pathMsg="This is the main file of PLTEC Converter", ErrorCode="NoError_This_is_a_log5").log()

    with open(path_for_output_file+".asm", "w") as f:
        f.write(Output_data)
        f.close()

    if object_:
        labels_addr = {}
        current_addr = 0

        Object = None
        if mode[0] == "elf" and mode[1] == "x86":
            Object = OBJECT.ELF(labels_addr, current_addr)

        section_t, section_d = Object.convert_to_o(Output_data, Data_SECTION)
        section_headers = Object.finalize_elf()
        if debug:
            Logger.LOG(TitleMsg="Section Headers", errorTitle="LOG", debugMsg=str(section_headers), Message="These are the section headers of the object output.", descriptionMsg="The section-headers/string-table is/are a specific portion inside the object file that define the location/address of every section in the machine code."
                    , path="PLTEC/main.py", pathMsg="This is the main file of PLTEC Converter", ErrorCode="NoError_This_is_a_log5").log()
        headers = Object.create_header(section_headers, debug)

        O = headers + section_headers + section_t + section_d

        if debug:
            Logger.LOG(TitleMsg="Object Headers", errorTitle="LOG", debugMsg=str(headers), Message="These are the headers of the object file.", descriptionMsg="The headers inside the object file are one of the most important elements that enable the OS/program to understand about the type/get the info about the file and code."
                    , path="PLTEC/main.py", pathMsg="This is the main file of PLTEC Converter", ErrorCode="NoError_This_is_a_log6").log()
            Logger.LOG(TitleMsg="Object Full Code", errorTitle="LOG", debugMsg=str(O), Message="This is the full Object File Code.", descriptionMsg="A object file is a source file that is created before every file is linked together to form a executable file like (elf32, elf64, exe32, exe64, etc). A object file can be reffered to as a full runnable file but it has to be linked before it can be executed and hence, object files are not executable."
                    , path="PLTEC/main.py", pathMsg="This is the main file of PLTEC Converter", ErrorCode="NoError_This_is_a_log7").log()

        with open(path_for_output_file+".o", "wb") as f:
            f.write(O)
            f.close()

if __name__ == "__main__":
    job()


import sys
import os
from typing import *

from . import *

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)

from phardwareitk.Extensions import HyperIn, HyperOut
from phardwareitk import Extensions

class Cursor:
    """A Class dedicated to cursor operations in command line.
    """
    @staticmethod
    def MoveCursorX(x: int) -> None:
        """Moves the cursor to the specified X postion.

        Args:
            x (int): X coordinate.
        """
        sys.stdout.write(f"\033[{x}G")
        sys.stdout.flush()
    @staticmethod
    def MoveCursor(x:int, y:int) -> None:
        """Move Cursor to specified position.

        Args:
            x (int): The X coordinate.
            y (int): the Y coordinate.
        """
        sys.stdout.write(f'\033[{y};{x}H')
        sys.stdout.flush()

    @staticmethod
    def HideCursor() -> None:
        """Hide the cursor."""
        sys.stdout.write('\033[?25l')
        sys.stdout.flush()

    @staticmethod
    def ShowCusor() -> None:
        """Show the cursor."""
        sys.stdout.write('\033[?25h')
        sys.stdout.flush()

    @staticmethod
    def MoveCursorUp(n: int) -> None:
        """Move cursor up by n lines."""
        if n == 0:
            return None
        if not n:
            return None
        sys.stdout.write(f'\033[{n}A')
        sys.stdout.flush()

    @staticmethod
    def MoveCursorDown(n: int) -> None:
        """Move cursor down by n lines."""
        if n == 0:
            return None
        if not n:
            return None
        sys.stdout.write(f'\033[{n}B')
        sys.stdout.flush()

    @staticmethod
    def MoveCursorRight(n: int) -> None:
        """Move cursor right by n columns."""
        if n == 0:
            return None
        if not n:
            return None
        sys.stdout.write(f'\033[{n}C')
        sys.stdout.flush()

    @staticmethod
    def MoveCursorLeft(n: int) -> None:
        """Move cursor left by n columns."""
        if n == 0:
            return None
        if not n:
            return None
        sys.stdout.write(f'\033[{n}D')
        sys.stdout.flush()

    @staticmethod
    def SaveCursorPosition() -> None:
        """Save the current cursor position."""
        sys.stdout.write('\033[s')
        sys.stdout.flush()

    @staticmethod
    def RestoreCursorPosition() -> None:
        """Restore the cursor to its last saved position."""
        sys.stdout.write('\033[u')
        sys.stdout.flush()

    @staticmethod
    def SetCursorToBeginningOfLine() -> None:
        """Move the cursor to the beginning of the current line."""
        sys.stdout.write('\033[0G')
        sys.stdout.flush()

    @staticmethod
    def SetCursorToEndOfLine() -> None:
        """Move the cursor to the end of the current line."""
        sys.stdout.write('\033[999C')  # Effectively moves to the end of the line
        sys.stdout.flush()

    @staticmethod
    def HideCursorTemporarily() -> None:
        """Temporarily hide the cursor (until next action)."""
        sys.stdout.write('\033[?25l')
        sys.stdout.flush()

    @staticmethod
    def ShowCursorTemporarily() -> None:
        """Temporarily show the cursor again after it was hidden."""
        sys.stdout.write('\033[?25h')
        sys.stdout.flush()

    @staticmethod
    def SetCursorPositionToHome() -> None:
        """Set the cursor to the top left corner (1, 1)."""
        sys.stdout.write('\033[H')
        sys.stdout.flush()

    @staticmethod
    def SetCursorPositionToEnd() -> None:
        """Set the cursor to the bottom right corner of the terminal."""
        sys.stdout.write('\033[999;999H')  # Moves to the bottom-right of the screen
        sys.stdout.flush()

    @staticmethod
    def MoveToNextWord() -> None:
        """Move the cursor to the next word."""
        sys.stdout.write('\033[W')
        sys.stdout.flush()

    @staticmethod
    def MoveToPreviousWord() -> None:
        """Move the cursor to the previous word."""
        sys.stdout.write('\033[W')
        sys.stdout.flush()

    @staticmethod
    def MoveCursorToTop() -> None:
        """Move the cursor to the top of the terminal."""
        sys.stdout.write('\033[H')
        sys.stdout.flush()

    @staticmethod
    def MoveCursorToBottom() -> None:
        """Move the cursor to the bottom of the terminal."""
        sys.stdout.write('\033[999;999H')  # Effective way to scroll down
        sys.stdout.flush()

    @staticmethod
    def SetCursorVisibility(is_visible: bool) -> None:
        """Set the cursor visibility."""
        sys.stdout.write('\033[?25h' if is_visible else '\033[?25l')
        sys.stdout.flush()

    @staticmethod
    def CurrentCursorPosition() -> tuple:
        """Get the current cursor position using BLOCKING. (row(y), column(x))."""
        sys.stdout.write('\033[6n')  # Request cursor position
        sys.stdout.flush()
        
        cursor_position = sys.stdin.read(6)  # Read the response
        if cursor_position.startswith('\033['):
            y, x = cursor_position[2:-1].split(';')  # Format: '\033[y;xR'
            return (int(y), int(x))
        return (0, 0)  # Default if no valid position is returned

class Screen:
    @staticmethod
    def ClearCurrentLine() -> None:
        """Clear the current line."""
        sys.stdout.write('\033[2K')
        sys.stdout.flush()

    @staticmethod
    def ClearLine(y: int) -> None:
        """Clear a specified line."""
        sys.stdout.write(f'\033[{y};1H\033[2K')
        sys.stdout.flush()

    @staticmethod
    def ClearScreen() -> None:
        """Clear the entire screen."""
        sys.stdout.write('\033[2J')
        sys.stdout.flush()

    @staticmethod
    def ClearScreenFromCursorDown() -> None:
        """Clear the screen from the cursor's current position to the bottom."""
        sys.stdout.write('\033[J')
        sys.stdout.flush()

    @staticmethod
    def ClearScreenFromCursorUp() -> None:
        """Clear the screen from the cursor's current position to the top."""
        sys.stdout.write('\033[1J')
        sys.stdout.flush()

    @staticmethod
    def SetScreenBackgroundColor(color: str) -> None:
        """Set the background color of the terminal."""
        sys.stdout.write(f'\033[{color}m')
        sys.stdout.flush()

    @staticmethod
    def SetScreenForegroundColor(color: str) -> None:
        """Set the foreground color of the terminal."""
        sys.stdout.write(f'\033[{color}m')
        sys.stdout.flush()

    @staticmethod
    def SetScreenColorReset() -> None:
        """Reset the screen's colors to default."""
        sys.stdout.write('\033[0m')
        sys.stdout.flush()

    @staticmethod
    def HideScreenCursor() -> None:
        """Hide the terminal cursor."""
        sys.stdout.write('\033[?25l')
        sys.stdout.flush()

    @staticmethod
    def ShowScreenCursor() -> None:
        """Show the terminal cursor."""
        sys.stdout.write('\033[?25h')
        sys.stdout.flush()

    @staticmethod
    def GetTerminalSize() -> tuple:
        """Get the current terminal size (rows, columns)."""
        size = os.popen('stty size', 'r').read().split()
        return (int(size[0]), int(size[1]))

    @staticmethod
    def SetTerminalSize(rows: int, cols: int) -> None:
        """Set the terminal size."""
        sys.stdout.write(f'\033[8;{rows};{cols}t')
        sys.stdout.flush()

    @staticmethod
    def EnableAlternateScreenBuffer() -> None:
        """Enable the alternate screen buffer."""
        sys.stdout.write('\033[?1049h')
        sys.stdout.flush()

    @staticmethod
    def DisableAlternateScreenBuffer() -> None:
        """Disable the alternate screen buffer."""
        sys.stdout.write('\033[?1049l')
        sys.stdout.flush()

    @staticmethod
    def ScrollUp(n: int) -> None:
        """Scroll up n lines in the terminal."""
        if n == 0:
            return None
        if not n:
            return None
        sys.stdout.write(f'\033[{n}S')
        sys.stdout.flush()

    @staticmethod
    def ScrollDown(n: int) -> None:
        """Scroll down n lines in the terminal."""
        if n == 0:
            return None
        if not n:
            return None
        sys.stdout.write(f'\033[{n}T')
        sys.stdout.flush()

    @staticmethod
    def ClearTabs() -> None:
        """Clear all tab stops in the terminal."""
        sys.stdout.write('\033[3g')
        sys.stdout.flush()

    @staticmethod
    def SetTab() -> None:
        """Set a tab stop at the current cursor position."""
        sys.stdout.write('\033[H')
        sys.stdout.flush()

    @staticmethod
    def ResetTerminal() -> None:
        """Reset the terminal to its default state."""
        sys.stdout.write('\033c')
        sys.stdout.flush()

    @staticmethod
    def SetCursorStyle(style: int) -> None:
        """Set the cursor style (0: block, 1: underline, 2: bar)."""
        sys.stdout.write(f'\033[{style} q')
        sys.stdout.flush()

    @staticmethod
    def EnableRawMode() -> None:
        """Enable raw mode (no input processing)."""
        sys.stdout.write('\033[?1049h')  # Use alternate buffer
        sys.stdout.flush()

    @staticmethod
    def DisableRawMode() -> None:
        """Disable raw mode."""
        sys.stdout.write('\033[?1049l')  # Return to normal buffer
        sys.stdout.flush()

    @staticmethod
    def ClearFromCursorToTop() -> None:
        """Clear The Screen from the cursor to the top of the screen.
        """
        sys.stdout.write('\033[1J')
        sys.stdout.flush()

    @staticmethod
    def ClearFromCursorToBottom() -> None:
        """Clear The Screen from the cursor to the bottom of the screen
        """
        sys.stdout.write('\033[J')
        sys.stdout.flush()

    @staticmethod
    def ClearFromCursorToEndOfLine() -> None:
        """Clear The Screen from the cursor to the end of the line
        """
        sys.stdout.write('\033[K')
        sys.stdout.flush()

    @staticmethod
    def ClearCharacter(x: int, y: int) -> None:
        """Clear the character at the specified location (x, y)."""
        Cursor.MoveCursor(x, y)
        sys.stdout.write(' ')  # Write a space to clear the character
        sys.stdout.flush()

class Text:
    @staticmethod
    def WriteText(*values:object, seperator:Union[str, None]=" ", endl:Union[str, None]="", File:Union[str, None]=None, Flush:bool=False, backgroundColorEnabled:bool=False, FontEnabled:bool=False, Font:Extensions.TextFont=Extensions.TextFont()) -> None:
        """Writes the specified text. (USES phardwareitk.Extensions.HyperOut.printH)

        Args:
            seperator (Union[str, None], optional): Seperator between values in [*values]. Defaults to " ".
            endl (Union[str, None], optional): The ending string to include at the end of the values. None is don't want. Defaults to "\n".
            File (Union[str, None], optional): Wether to save the print contents in a file before printing in console. Defaults to None.
            Flush (bool, optional): Wether to use Flush during printing. Defaults to False.
            backgroundColorEnabled (bool, optional): Wether to enable background-color. Defaults to None.
            FontEnabled (bool, optional): Wether to enable custom font. Defaults to False.
            Font (Extensions.TextFont, optional): The font. Defaults to Extensions.TextFont().
        """
        HyperOut.printH(*values, seperator=seperator, endl=endl, File=File, Flush=Flush, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)

    def InputText(*values:object, seperator:Union[str, None]=" ", endl:Union[str, None]="\n", backgroundColorEnabled:bool=False, FontEnabled:bool=False, Font:Extensions.TextFont=Extensions.TextFont()) -> str:
        """Returns the entered user data from input. (USES phardwareitk.Extensions.HyperIn.inputH)

        Args:
            seperator (Union[str, None], optional): Seperator between values in [*values]. Defaults to " ".
            endl (Union[str, None], optional): The ending string to include at the end of the values. None is don't want. Defaults to "\n".
            backgroundColorEnabled (bool, optional): Wether to enable background-color. Defaults to False.
            FontEnabled (bool, optional): Wether to enable custom font. Defaults to False.
            Font (Extensions.TextFont, optional): The font. Defaults to Extensions.TextFont().
        
        Returns:
            str: The user inputted string.
        """
        return HyperIn.inputH(*values, seperator=seperator, endl=endl, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)
    
    @staticmethod
    def DeleteChar() -> None:
        """Simulates Delete Key in terminal.
        """
        sys.stdout.write("\033[3~")
        sys.stdout.flush()

    @staticmethod
    def BackSpaceChar() -> None:
        """Simulates Backspace in terminal.
        """
        sys.stdout.write("\b")
        sys.stdout.write(" ")
        sys.stdout.write("\b")
        sys.stdout.flush()


from typing import *

import os

class Requirements:
    @staticmethod
    def Modules(mode:Union[str, None, list]=None) -> Union[str, list, None]:
        """
        Args:
            mode (Union[str, None, list]): This argument allows the program to understand the specified mode. modes ->
                str: Returns the string representation of the required modules seperated via ' '.
                list: Returns the list representation of the required modules.
                None: Downloads the required modules.

        Returns:
            Union[str, list, None]: Returned Type will be the mode type.
        """
        download:bool = False
        list_:bool = False
        
        if mode == None:
            download = True
        else:
            download = False
        
        if isinstance(mode, str):
            list_ = False
        else:
            list_ = True

        required_modules = "cryptography psutil keyboard cython PyOpenGL pysdl2%pysdl2-dll"
        logPLUSstr = "Modules That are optional for downloading ->\n1. PheonixAppAPI: Great Module and Provides almost all functionality required for a devloper to move on. From playing games to working with Files.\nDetails ->\n\tSpace: MAX 50mb for the module + 6 GB for building PyQt5 [Microsoft C++ Build Tools 14 or higher] (you can skip the 6 GB by downloading a premade PyQt5 whl file online)."
        required_modules_list = []
        optional_required_compiling = "Please, to use GUI/HGame (Screen_X and Draw_X class) compile the renderGUI.pyx file using gcc (Microsoft Build tools/Mingw64)."

        if required_modules == "":
            print("No required modules.\n"+logPLUSstr+f"\n{optional_required_compiling}")
            return None

        if not " " in required_modules:
            required_modules_list.append(required_modules)
        else:
            required_modules_list = required_modules.split(" ")

        if download:
            for i, v in enumerate(required_modules_list):
                if "%" in v:
                    v = v.split("%")

                v_:str = v
                
                for k, j in enumerate(v):
                    v_ += j
                
                v = v_

                os.system(f"pip install {v}")

            print(logPLUSstr+"\n"+optional_required_compiling)

            return None
        
        if list_:
            return required_modules_list
        else:
            return required_modules
        
if __name__ == "__main__":
    opr:str = input("Usage (modules): ")
    opr = opr.lower()

    if opr == "modules":
        Requirements.Modules(None)
    else:
        print("Unknown Usage -> [" + opr + "]")


"""Hardware Hyper Game (HGame).

Base Functions:
    Initialize: The first function to be called before any other.

Classes:
    Screen: Classic Screen Class. Uses Tkinter. (Fast but not very fast)
    Draw: Classic Draw Class. Uses Tkinter. (Fast but not very fast)
    Screen_ADV: Advanced Screen Class. Uses PHardwareGUI (Very Fast and efficient. Recommended for Complex apps)
    Draw_ADV: Advanced Draw Class. Uses PHardwareGUI (Very Fast and efficient. Recommended for Complex apps)
    Screen_X: Expert Level Screen Class. Uses OpenGL (Provides Low level interactions and hence is the fastest. Recommended for Very complex apps and 3D apps).
    Draw_X: Expert Level Screen Class. Uses OpenGL (Provides Low level interactions and hence is the fastest. Recommended for Very complex apps and 3D apps).
"""

import tkinter
import os
import sys

PHardwareITK = len(sys.path) - 1
PHardwareITK_P = os.path.join(os.path.dirname(__file__), "..", "..")

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)

import phardwareitk.LIB
from phardwareitk.CLI import cliToolKit
from phardwareitk.ErrorSystem import ErrorSystem as es

from phardwareitk.Extensions.HyperOut import *

global Debug
global Stdout
global Stdin
global OpenGl
global Gd

Debug:bool = False
Stdout:str = ""
Stdin:str = ""
Stdout_Stdin_list:list = ["console", "ext_window", "window"]

OpenGl:bool = False

Gd:str = ""

G_dictList:list = ["2d", "3d"]

def Initialize(debug_stdout:bool=False, stdout:str="console", stdin:str="window", G_2D_3D:int=0, G_dict:dict[0,str]={0: "2d", 1: "3d"}, OpenGL:bool=False) -> None:
    global Debug
    global Stdout
    global Stdin
    global Stdout_Stdin_list
    global OpenGl
    global Gd
    global G_dictList
    
    Debug = debug_stdout
    if stdout in Stdout_Stdin_list:
        Stdout = stdout
    else:
        es.PheonixArgumentError(f"Stdout must be specified or must be present in the list of stdout-compatible streams. LISTS [{Stdout_Stdin_list}]")
    if stdin in Stdout_Stdin_list:
        Stdin = stdin
    else:
        es.PheonixArgumentError(f"Stdin must be specified or must be present in the list of the stdin-compatible streams. LISTS [{Stdout_Stdin_list}]")
    
    if G_2D_3D in G_dict.keys():
        if G_dict[G_2D_3D].lower() in G_dictList:
            Gd = G_dict[G_2D_3D].lower()
        else:
            es.PheonixArgumentError(f"G_dict values must be supported. LISTS [{G_dictList}]")
    else:
        es.PheonixArgumentError(f"G_2d_3D must be specified or must be present in the G_dict. G_dict -> [{G_dict}]")

    OpenGl = OpenGL

    if Debug and Stdout == "console":
        printH(f"\n HGame Settings -> \n debug-stdout: {debug_stdout} \n stdout: {stdout} \n stdin: {stdin} \n Mode: {Gd} \n OpenGL: {OpenGl} ", FontEnabled=True, backgroundColorEnabled=True, Font=TextFont(font="Futura", font_color=Color("black"), font_background_color=Color("red")))

    return None

class Screen():
    """Screen class that controls the screen and is the first class to be used after initialization.
    NOTE: Uses [tkinter] and not [PHardware GUI] to boost performance as [tkinter] is a module developed only for this purpose wheras [PHardwareITK] is a module that provides a hardware interface toolkit and other functions.
    """
    global Debug
    global Stdout
    global Stdin
    global Stdout_Stdin_list
    global OpenGl
    global Gd
    global G_dictList

    def __init__(self, name:str="HGame", width:int=800, height:int=400, x:int=0, y:int=0, backgroundColor:Color=Color(), icon:str="") -> None:
        self.name:str = name
        self.width:int = width
        self.height:int = height
        self.x:int = x
        self.y:int = y
        self.backgroundColor:Color = backgroundColor

        self.icon:str = icon

        if Debug == True and Stdout == "console":
            printH(f"\nCreating Window:\n1.Name: {name}\n2.Width: {width}\n3.Height: {height}\n4.X: {x}\n5.Y: Y\n6.Background Color: {backgroundColor.color}\n")

        self.root:tkinter.Tk = tkinter.Tk()

        if icon:
            self.root.iconphoto(False, icon)

        self.root.title(name)
        self.root.geometry(f"{width}x{height}")

        self.grid:bool = False

    def GridMode(self) -> None:
        """Initializes the grid mode.
        """
        self.grid = True
        return None

    def Loop(self) -> None:
        """Creates the game loop.
        """
        self.root.mainloop()

class Draw:
    def __init__(self, screen:Screen) -> None:
        self.root = screen.root
        self.screen = screen


import os
import sys

PHardwareITK = len(sys.path) - 1
PHardwareITK_P = os.path.join(os.path.dirname(__file__), "..", "..", "..")

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)

from phardwareitk.PLTEC import Checker

Checker.PLTEC_initCheck()


from . import *
from typing import *
import time

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)



def inputH(*values:object, seperator:Union[str, None]=" ", endl:Union[str, None]=None, backgroundColorEnabled:bool=False, FontEnabled:bool=False, Font:TextFont=TextFont(), mask:bool=False, maskCharacter:str="*", History:Optional[list]=None, Regex_ReEntry:Optional[str]=None, Regex_Msg: Optional[str]=None, Regex_Time: Optional[float]=None, DefaultVal:Optional[str]=None, cpuHoggin:float=0.005) -> str:
    """Hyper version of Python's input function.

    
    NOTE: This function uses custom key input detection and, if any bug occurs, please post in [https://github.com/AkshuDev/PHardwareITK/issues]

    
    Args:
        seperator (Union[str, None], optional): The seperator between [*values]. Defaults to " ".
        endl (Union[str, None], optional): The string to add at the end of [*values]. None if don't want. Defaults to None.
        backgroundColorEnabled (bool, optional): Wether to enable background-color. Defaults to False.
        FontEnabled (bool, optional): Wether to enable the custom font. Defaults to False.
        Font (TextFont, optional): The font. Defaults to TextFont().
        mask (bool, optional): Whether to mask the input text. Defaults to False.
        maskCharacter (str, optional): The character used to mask the input text. Defaults to '*'.
        History (Optional[list], optional): If not None, Upon receiving arrow keys, The old prompts/History will appear (User choice to enter history or not). Defaults to None.
        Regex_ReEntry (Optional[str], optional): If not None, The input will be going through a validation via the given regex. Defaults to None.
        Regex_Msg (Optional[str], optional): If not None and if the Regex_ReEntry is not None, then if match fails, then this msg will be printed, without newline. Please add newline before if you want. Defaults to None.
        Regex_Time (Optional[float], optional): If not None and if the Regex_ReEntry is not None, then if match fails, then this specifies the time, the message will be shown for. Defaults to None.
        DefaultVal (Optional[str], optional): If not None, If the input is empty, the provided Value will be returned.
        cpuHoggin (float, optional): Cpu Hogging is very much required for programs like this. This is a complex input function, so unlike python's input function, it cannot avoid CPU usage. To prevent it a small delay is added at the end of the loop. IT CANNOT BE 0 as it can harm the CPU. Defaults to 0.005 seconds or 5 milliseconds.

    Returns:
        str: The user provided input
    """
    import string
    from phardwareitk.CLI import cliToolKit
    if not endl:
        endl = ""
    
    if Regex_ReEntry:
        import re

    final_Str = ""

    for i, val in enumerate(values):
        if i == 0:
            final_Str += val
        elif seperator == "" or not seperator:
            final_Str += val
        else:
            final_Str += seperator
            final_Str += val
    
    if endl and not endl == "":
        final_Str += endl

    if cpuHoggin < 0.003:
        cpuHoggin = 0.003

    cliToolKit.Text.WriteText(final_Str, Flush=True, seperator=seperator, endl=endl, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)

    buffer = ""
    bufferL:list = []
    key_pressed:bool = False
    key_pressedB:bool = False
    key_pressedHis: bool = False
    key_pressedL:bool = False
    key_pressedR:bool = False

    chars:int = 0
    cPrompt:int = 0
    cPos:int = 0

    bufferPrompt:str = ""

    import keyboard

    while True:
        key_ = keyboard.read_event(suppress=True)
        if key_.event_type == keyboard.KEY_DOWN:
            if key_.name in string.printable:
                if len(buffer) > 0:
                    buffer = buffer[:cPos]
                    buffer += key_.name + bufferPrompt[cPos:]
                else:
                    buffer = key_.name
                bufferPrompt = buffer
                if len(bufferL) > 0:
                    bufferL.insert(cPos + 1, key_.name)
                else:
                    bufferL.append(key_.name)
                chars += 1
                cPos += 1
                if not mask:
                    cliToolKit.Text.WriteText(key_.name, Flush=True, endl=None, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)
                    if buffer[cPos:] != "":
                        cliToolKit.Text.WriteText(buffer[cPos:], Flush=True, endl=None, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)
                        cliToolKit.Cursor.MoveCursorLeft(len(buffer[cPos:]))
                else:
                    cliToolKit.Text.WriteText(maskCharacter, Flush=True, endl=None, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)
                    if buffer[cPos:] != "": 
                        cliToolKit.Text.WriteText(maskCharacter*len(buffer[cPos:]), Flush=True, endl=None, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)
                        cliToolKit.Cursor.MoveCursorLeft(len(buffer[cPos:]))
            
            elif key_.name == "enter":
                if not key_pressed:
                    key_pressed = True
                    if not Regex_ReEntry:
                        print("\n", flush=True)
                        keyboard.unhook_all()
                        return buffer
                    else:
                        result = re.match(Regex_ReEntry, buffer)
                        if result:
                            if DefaultVal and (DefaultVal != "" and buffer == ""):
                                buffer = DefaultVal
                            print("\n", flush=True)
                            keyboard.unhook_all()
                            return buffer
                        else:
                            if not Regex_Msg or Regex_Msg == "":
                                continue
                            cliToolKit.Text.WriteText(Regex_Msg, Flush=True, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)
                            cliToolKit.Cursor.MoveCursorLeft(len(Regex_Msg))
                            if Regex_Time and Regex_Time > 0:
                                time.sleep(Regex_Time)
                            cliToolKit.Text.WriteText(" "*len(Regex_Msg))
                            cliToolKit.Cursor.MoveCursorLeft(len(Regex_Msg))
                            continue
            elif key_.name == "backspace":
                if not (key_pressedB and buffer != "") and (cPos > 0 and chars > 0):
                    key_pressedB = True
                    bufferL.pop(cPos - 1)
                    buffer = "".join(bufferL)
                    bufferPrompt = buffer
                    # Remove old text
                    if chars == cPos:
                        cliToolKit.Cursor.MoveCursorLeft(chars)
                    else:
                        cliToolKit.Cursor.MoveCursorLeft(cPos)
                    cliToolKit.Text.WriteText(" "*chars, FontEnabled=FontEnabled, Font=Font, Flush=True)
                    cliToolKit.Cursor.MoveCursorLeft(chars)
                    # Add new text
                    if not mask:
                        cliToolKit.Text.WriteText(buffer, FontEnabled=Font, backgroundColorEnabled=backgroundColorEnabled, Font=Font, Flush=True)
                    else:
                        cliToolKit.Text.WriteText(maskCharacter*(chars - 1), FontEnabled=Font, backgroundColorEnabled=backgroundColorEnabled, Font=Font, Flush=True)
                    chars -= 1
                    cPos -= 1
                    if not cPos == chars:
                        cliToolKit.Cursor.MoveCursorLeft(chars - cPos)
            elif key_.name == "space":
                if len(buffer) > 0:
                    buffer = buffer[:cPos]
                    buffer += " " + bufferPrompt[cPos:]
                else:
                    buffer = " "
                bufferPrompt = buffer
                if len(bufferL) > 0:
                    bufferL.insert(cPos + 1, " ")
                else:
                    bufferL.append(" ")
                chars += 1
                cPos += 1
                if not mask:
                    cliToolKit.Text.WriteText(" ", Flush=True, endl=None, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)
                    if buffer[cPos:] != "":
                        cliToolKit.Text.WriteText(buffer[cPos:], Flush=True, endl=None, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)
                        cliToolKit.Cursor.MoveCursorLeft(len(buffer[cPos:]))
                else:
                    cliToolKit.Text.WriteText(maskCharacter, Flush=True, endl=None, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)
                    if buffer[cPos:] != "": 
                        cliToolKit.Text.WriteText(maskCharacter*len(buffer[cPos:]), Flush=True, endl=None, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)
                        cliToolKit.Cursor.MoveCursorLeft(len(buffer[cPos:]))
            elif key_.name == "up":
                if History:
                    if not key_pressedHis:
                        try:
                            prompt = History[cPrompt]
                            buffer = prompt
                            if chars != 0:
                                if cPos == chars:
                                    cliToolKit.Cursor.MoveCursorLeft(chars)
                                else:
                                    cliToolKit.Cursor.MoveCursorLeft(cPos)
                                cliToolKit.Text.WriteText(" " * chars, Flush=True, FontEnabled=FontEnabled, Font=Font)
                                cliToolKit.Cursor.MoveCursorLeft(chars)
                            cliToolKit.Text.WriteText(prompt, Flush=True, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)
                            cPrompt += 1
                            chars = len(prompt)
                            cPos = chars
                        except IndexError:
                            cPrompt = 0
                            prompt = bufferPrompt
                            buffer = prompt
                            if cPos == chars:
                                cliToolKit.Cursor.MoveCursorLeft(chars)
                            else:
                                cliToolKit.Cursor.MoveCursorLeft(cPos)
                            cliToolKit.Text.WriteText(" " * chars, Flush=True, FontEnabled=FontEnabled, Font=Font)
                            cliToolKit.Cursor.MoveCursorLeft(chars)
                            cliToolKit.Text.WriteText(prompt, Flush=True, backgroundColorEnabled=backgroundColorEnabled, FontEnabled=FontEnabled, Font=Font)
                            chars = len(prompt)
                            cPos = chars
                        
                        key_pressedHis = True
                
            elif key_.name == "left":
                if not key_pressedL:
                    if cPos > 0:
                        cliToolKit.Cursor.MoveCursorLeft(1)
                        cPos -= 1
                    key_pressedL = True
            elif key_.name == "right":
                if not key_pressedR:
                    if cPos < chars:
                        cliToolKit.Cursor.MoveCursorRight(1)
                        cPos += 1
                    key_pressedR = True

        time.sleep(cpuHoggin) # Prevents hogging
        key_pressed = False
        key_pressedL = False
        key_pressedR = False
        key_pressedB = False
        key_pressedHis = False


"""
Under Development. The provided functions in gui_sdl and renderGUI are just for making the gui process simple.
You can use this file to use SDL and OpenGL functions to create whatever you want.

NOTES:
1. Uses SDL2 (Simple DirectMedia Layer) and OpenGL (Open Graphics Library)

2. Both SDL2 and OpenGL are cross-platform.

3. OpenGL may vary depending on graphics card.

4. To use OpenGL follow the steps below ->

 a. Change the directory to phardwareitk Folder.

 b. Do - '''pip install cython'''


 c. Do - '''cythonize -i GUI/renderGUI.pyx'''


 d. Download GCC


 e. Run - '''gcc -shared -pthread -fPIC -fwrapv -O2 -Wall -fno-strict-aliasing -I/[Path to Main Python Folder] -o GUI/renderGUI.[so in linux and pyd in windows] GUI/renderGUI.c'''
 
 
 f. The functions from renderGUI will be imported as OpenGUI.[functions]
 
5. SDL2 Only supports 2d and provides more simple functions. Whereas OpenGL is known for its complexity and performance. OpenGL directly communicates with the GPU and hence needs to be written in C. OpenGL supports both 2D and 3D and complex shapes.

6. Better to import phardwareitk.Extensions, for classes like PIcon, PWidget, Color, TextFont, it also provides HyperIn and HyperOut files, including better versions of python buitins"""

from . import *

from sdl2 import *
from OpenGL import *

import sdl2
import OpenGL.GL

from OpenGL.GL import *

import sys
import os

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)

from phardwareitk.GUI.gui_sdl import *
from phardwareitk import LIB

# sys.path.append(os.path.join(os.path.dirname(__file__), ".."))

from phardwareitk.ErrorSystem import ErrorSystem as Errors

if Unknown_os:
    Errors.PheonixOsError("Unknown OS detected. Program doesn't know will this GUI system work properly in this OS. To Test, please SET the Windows/Linux/Unix to True and Unknown OS to False in GUI.__init__.")

if os.path.exists(LIB.Paths.renderGUIPYD_GUI) or os.path.exists(LIB.Paths.renderGUISO_GUI):
    from phardwareitk.GUI import renderGUI as OpenGui

class ToStandaloneApplication():
    """Convert the GUI app to a standalone application. [.EXE/.ELF/...]
    """
    def __init__(self, getRequirements:bool, icon:Optional[str], file:str, windowed:bool=False) -> None:
        """Initialize function of ToStandaloneApplication.

        Args:
            getRequirements (bool): Download the requirements?
            icon (Optional[str]): The icon to be used. Leave for default.
            file (str): The main .py file.
            windowed (bool): Whether to run the app inside a window. NOTE: If the app is a GUI application then it is recommended to Set this to False.
        """
        self.gr = getRequirements
        self.file = file
        self.icon = icon
        self.windowed = windowed
        self.cmd = ""

        if self.gr:
            self.getReq()

        self.makeCmd()

        if self.windowed == True:
            self.addWIN()
        
        if (self.icon != "" and self.icon) and windowed:
            self.addIcon()

        self.runCmds1()
        self.MakeStandaloneApplication()

    def runCmds1(self):
        self.cmd += self.file

    def makeCmd(self):
        self.cmd += f"pyinstaller --onefile"

    def addWIN(self):
        self.cmd += f"--windowed "

    def addIcon(self):
        self.cmd += f"--icon={self.icon} "

    def getReq(self):
        os.system("pip install pyinstaller")

    def MakeStandaloneApplication(self):
        """Makes a standalone application.
        """
        os.system(self.cmd)


"""
HyperPython is a Extension in PHardwareITK created by one teen. (Pheonix Studios (Akshobhya)).
This converts Python code to C/ASM.

Working on -> Compiler

Next Update -> C to Executable without GCC/Clang/....


NOTES:

1. Under Development.

2. Made because I wanted to.

3. Difference between Cython and HyperPython is as follows.

3.a. Hyper Python generates a Pure C file that can't communitcate with python, but cython does.

3.b. Cython generates a .c file from .pyx that needs to be compiled via gcc/clang/.... whereas HyperPython generates a .c file from .py that needs to be compiled via gcc/clang/.... FOR NOW. HyperPython.ToExecutable class under development.

4. You can use this but remember, this is under development.

5. This might give a huge benifit of file size.
"""

MAXBUFFER = 1024
MAXLINES = 50000

from typing import *

class bltin:
    def __init__(self, *value):
        self.isbltin:bool = True
        self.value = value

    @property
    def buitlin(self):
        return self.isbltin
    
    def __str__(self):
        return str(self.value)

class HyperList:
    def __init__(self, list1:list, list2:list, *args:list) -> None:
        self.list1 = list1
        self.list2 = list2
        self.lists = args

        self.hyperList = self.MHyperList(list1, list2, self.lists)

    @property
    def list_(self) -> list[tuple[list]]:
        return self.hyperList
    
    def MHyperList(self, list1:list, list2:list, *lists:list) -> list[tuple[list]]:
        lists_ = [list1, list2]
        for list in lists:
            lists_.append(list)

        if len(lists_) % 2 == 0:
            pass
        else:
            lists_.append([])

        hyperList = []

        skipTo = -1
        for i, v in enumerate(lists_):
            if i != skipTo:
                tuple_ = (v, lists_[i + 1])
                skipTo = i + 1
                hyperList.append(tuple_)

        return hyperList

    def append(self, list1:list, list2:list):
        list_ = list1

        for i, v in enumerate(list2):
            list_.append(v)

        self.hyperList.append(tuple(list_))

class buffer(bltin):
    def __init__(self, *args) -> None:
        self.args:tuple = args
        self.len:int = None
        self.buffer_:list = None

        buffer_:list = []

        if self.args and len(self.args) >= 1:
            for i, v in enumerate(self.args):
                for char in v:
                    buffer_.append(char)

        self.len = len(buffer_)
        self.buffer_ = buffer_

        super().__init__(self.buffer_)

    def __list__(self):
        return str(self.buffer_)
    
    @property
    def value(self) -> list:
        return self.buffer_
    
    @property
    def length(self) -> int:
        return self.len
    
class Generator_C:
    def __init__(self, InfTyp:HyperList.list_, includeHeaders:list):
        self.InfTyp = InfTyp
        self.includeHeaders = includeHeaders

        self.code = "#include <stdlib.h>\n#include <stdio.h>\n"

        self.finalCode = ""

        self.funcs:dict = {}

        print(InfTyp)
        print(includeHeaders)

    def MakeMainFunc(self):
        self.code += "\nint main(int argc, char** argv){\n"
        self.funcs["main"] = ""

    def Generate(self):
        if len(self.includeHeaders) > 0:
            for header in self.includeHeaders:
                self.code += "#include <" + str(header) + ">\n"
        
        self.MakeMainFunc()

        beautify = "\t"

        for i, v in enumerate(self.InfTyp):
            type_, class_, name, value = v

            type_ = beautify + str(type_)

            finalVar = type_ + " " + str(name) + " = " + str(value) + ";\n"

            self.funcs["main"] += "\n" + finalVar

        self.code += self.funcs["main"] + "}\n"

        print(self.code)

class Tokenizer_C:
    def __init__(self, code:str):
        # Tokenizer
        self.code = code

    def tokenize(self):
        import tokenize
        from io import BytesIO


        return tokenize.tokenize(BytesIO(self.code.encode()).readline())
    
    def infer_types(self):
        import ast
        tree = ast.parse(self.code)

        InfrTyp:HyperList = HyperList([], [])

        includeHeaders:list = []

        for node in ast.walk(tree):
            if isinstance(node, ast.Assign):
                # Handle variable assignment
                for target in node.targets:
                    if isinstance(target, ast.Name):
                        var_name = target.id
                        value = node.value
                        var_type = "unknown"
                        var_typeClass = None

                        if isinstance(value, ast.Constant):
                            if isinstance(value.value, int): #int
                                if str(value.value) == "True" or str(value.value) == "False":
                                    var_type = "bool"
                                    var_typeClass = bool
                                    InfrTyp.append([var_type, var_typeClass], [var_name, str(value.value).lower()])
                                    includeHeaders.append("stdbool.h")
                                else: 
                                    var_type = "int"
                                    var_typeClass = int
                                    InfrTyp.append([var_type, var_typeClass], [var_name, value.value])
                            elif isinstance(value.value, float): # float
                                var_type = "float"
                                var_typeClass = float
                                InfrTyp.append([var_type, var_typeClass], [var_name, value.value])
                            elif isinstance(value.value, str): # String
                                if len(value.value) >= 1:
                                    if len(value.value) == 1:
                                        var_type = "char"
                                        var_typeClass = buffer
                                        InfrTyp.append([var_type, var_typeClass], [var_name, value.value])
                                    else:
                                        var_type = "char*"
                                        var_typeClass = str
                                        InfrTyp.append([var_type, var_typeClass], [var_name, value.value])
                                else:
                                    var_type = "null"
                                    var_typeClass = None
                                    InfrTyp.append([var_type, var_typeClass], [var_name, value.value])
                            elif isinstance(value.value, bool): # Bool
                                var_type = "bool"
                                var_typeClass = bool
                                InfrTyp.append([var_type, var_typeClass], [var_name, str(value.value).lower()])
                                includeHeaders.append("stdbool.h")

        return InfrTyp.list_, includeHeaders
                            

# 0 -> C
# 1 -> ASM
# 2 -> EXE
# 3 -> ELF
# 4 -> AEF
# 5 -> CAEF
def Convert(to:int=0, code:str="", file:str="") -> int:
    """_summary_

    Args:
        to (int, optional): Which type to convert to. Defaults to 0.
        Available Types ->
            # 0 -> C
            # 1 -> ASM
            # 2 -> EXE
            # 3 -> ELF
            # 4 -> AEF
            # 5 -> CAEF

        code (str, optional): If code is in string form, pass here.
        file (str, optional): If code in inside a file, pass file path here.

    Returns:
        int: 0 for success, 1 for failure, -11 for OSError.
    """
    if code and code != "":
        if to == 0:
            tokenizer = Tokenizer_C(code)
            tokens = tokenizer.tokenize()
            print(tokens)
            ifr, headers = tokenizer.infer_types()
            ifr.pop(0)
            ifr.pop(0)
            generator = Generator_C(ifr, headers)
            generator.Generate()
    else:
        if not file or file == "":
            raise ValueError("Neither File nor Code Arguments have been passes")
    
        else:
            code = ""
            with open(file, "r") as f:
                code = f.read()
                f.close()

            if to == 0:
                tokenizer = Tokenizer_C(code)
                tokens = tokenizer.tokenize()
                ifr, headers = tokenizer.infer_types(tokens)

                ifr.pop(0)
                ifr.pop(0)

                generator = Generator_C(ifr, headers)
                generator.Generate()
    return 0


import os
import sys
import datetime
from typing import *

class LOG():
    def __init__(self, dateTime:bool=True, logLevel:int=1, debugLog:bool=True, pathInclude:bool=True, pathMsgInclude:bool=True, MessageInclude:bool=True,
                descriptionInclude:bool=True, errorTitleInclude:bool=True, titleInclude:bool=True, TitleMsg:str = "", errorTitle:str="", debugMsg:str="", dateFormat:str="%d-%m-%y %I:%M:%S %p",
                path:str="", pathMsg:str="", Message:str="", descriptionMsg:str="", ErrorCodeInclude:bool=True, ErrorCode: str="",
                LogFormat:str="\n$dateTime$:%$Title$%<->%$Error->%$errorTitle$:\n$Message$\n\nDescription:%$description$\n\nPath:%$path$:%$pathMsg$\n\nDebug:%$debugLog$\nCode:%$ErrorCode$^PERCENT^ ") -> None:
        
        """
        MAIN initialize function


        LogFormat % refers to the the sperater string after ^PERCENT^. DEFAULT: SPACE (' ')


        VARIABLES IN LogFormat: ['dateTime', 'Title', 'errorTitle', 'Message', 'description', 'path', 'pathMsg', 'debugLog', 'ErrorCode']
        """

        dateNOW = datetime.datetime.now()
        self.dateTime = dateNOW.strftime(dateFormat)

        self.logLevel:bool = logLevel
        self.errorCodeInclude = ErrorCodeInclude
        self.debugLog:bool = debugLog
        self.pathInclude:bool = pathInclude
        self.pathMsgInclude:bool = pathMsgInclude
        self.MessageInclude:str = MessageInclude
        self.descriptionInclude:bool = descriptionInclude
        self.errorTitleInclude:bool = errorTitleInclude
        self.titleInclude:bool = titleInclude
        self.titleMsg:str = TitleMsg
        self.errorTitle:str = errorTitle
        self.debugMsg:str = debugMsg
        self.dateFormat:str = dateFormat
        self.path:str = path
        self.pathMsg:str = pathMsg
        self.Message:str = Message
        self.descriptionMsg:str = descriptionMsg
        self.LogFormat:str = LogFormat
        self.errorCode:str= ErrorCode

        self.LOGvariables:dict = {
            'dateTime': self.dateTime,
            'Title': self.titleMsg,
            'errorTitle': self.errorTitle, 
            'Message': self.Message, 
            'description': self.descriptionMsg, 
            'path': self.path, 
            'pathMsg': self.pathMsg, 
            'debugLog': self.debugMsg, 
            'ErrorCode': self.errorCode,
            "dateTime_include": dateTime,
            "Title_include": self.titleInclude,
            "errorTitle_include": self.errorTitleInclude,
            "Message_include": self.MessageInclude,
            "description_include": self.descriptionInclude,
            "path_include": self.pathInclude,
            "pathMsg_include": self.pathMsgInclude,
            "debugLog_include": self.debugLog,
            "ErrorCode_include": self.errorCodeInclude,
        }

    def _compileLOG(self) -> str:
        """
        Not to be called directly from script.
        """
        log:str = ""

        logSplit = self.LogFormat.split("$")
        logSeperator = logSplit[len(logSplit)-1].split("^PERCENT^")[1]

        logSplit.pop(len(logSplit)-1)

        for k, j in enumerate(logSplit):
            logSplit[k] = j.replace("%", logSeperator)

        for i, v in enumerate(logSplit):
            if v in self.LOGvariables.keys():
                if self.LOGvariables[v+"_include"] == True:
                    log += self.LOGvariables[v]
            else:
                log += v

        return log
    
    def log(self, returnType:Union[str, None]=None) -> Union[str, None]:
        """
        'None' means printing the log and 'str' means, it returns the log.
        """
        log:str = self._compileLOG()

        if returnType is None:
            print(log)
            return None
        else:
            return log



from typing import *

import os
import sys

PHardwareITK = len(sys.path) - 1
PHardwareITK_P = os.path.join(os.path.dirname(__file__), "..", "..")

class Color:
    """Color Class for defining color.
    """
    def __init__(self, color:str="white", alpha:int=255) -> None:
        """Init of color class.

        Args:
            color (str, optional): The color, leave and chose a function if, want to use RGB. Defaults to "white".
        """
        self.color_names_rgb:dict[str, tuple[int, int, int]] = {
            "red": (255, 0, 0),
            "green": (0, 255, 0),
            "blue": (0, 0, 255),
            "yellow": (255, 255, 0),
            "cyan": (0, 255, 255),
            "magenta": (255, 0, 255),
            "black": (0, 0, 0),
            "white": (255, 255, 255),
            "gray": (169, 169, 169),
            "grey": (169, 169, 169),
            "silver": (192, 192, 192),
            "maroon": (128, 0, 0),
            "olive": (128, 128, 0),
            "lime": (0, 255, 0),
            "aqua": (0, 255, 255),
            "fuchsia": (255, 0, 255),
            "teal": (0, 128, 128),
            "navy": (0, 0, 128),
            "purple": (128, 0, 128),
            "pink": (255, 192, 203),
            "brown": (139, 69, 19),
            "beige": (245, 245, 220),
            "chocolate": (210, 105, 30),
            "coral": (255, 127, 80),
            "indigo": (75, 0, 130),
            "lavender": (230, 230, 250),
            "violet": (238, 130, 238),
            "peach": (255, 218, 185),
            "gold": (255, 215, 0),
            "plum": (221, 160, 221),
            "orchid": (218, 112, 214),
            "crimson": (220, 20, 60),
            "chartreuse": (127, 255, 0),
            "tomato": (255, 99, 71),
            "salmon": (250, 128, 114),
            "seashell": (255, 245, 238),
            "blush": (222, 93, 131),
            "wheat": (245, 222, 179),
            "khaki": (240, 230, 140),
            "mint": (189, 252, 201),
            "forest": (34, 139, 34),
            "lemon": (255, 247, 0),
            "pearl": (234, 234, 234),
            "periwinkle": (204, 204, 255),
            "sand": (194, 178, 128),
            "emerald": (80, 200, 120),
            "russet": (128, 70, 27),
            "scarlet": (255, 36, 0),
            "charcoal": (54, 69, 79),
            "rose": (255, 99, 123),
            "tangerine": (255, 129, 0),
            "bordeaux": (127, 0, 24)
        }

        self.hex_codes_rgb:dict[str, tuple[int, int, int]] = {
            "#ff0000": (255, 0, 0),
            "#00ff00": (0, 255, 0),
            "#0000ff": (0, 0, 255),
            "#ffff00": (255, 255, 0),
            "#00ffff": (0, 255, 255),
            "#ff00ff": (255, 0, 255),
            "#000000": (0, 0, 0),
            "#ffffff": (255, 255, 255),
            "#a9a9a9": (169, 169, 169),
            "#c0c0c0": (192, 192, 192),
            "#800000": (128, 0, 0),
            "#808000": (128, 128, 0),
            "#00ff00": (0, 255, 0),
            "#00ffff": (0, 255, 255),
            "#ff00ff": (255, 0, 255),
            "#008080": (0, 128, 128),
            "#000080": (0, 0, 128),
            "#800080": (128, 0, 128),
            "#ffc0cb": (255, 192, 203),
            "#8b4513": (139, 69, 19),
            "#f5f5dc": (245, 245, 220),
            "#d2691e": (210, 105, 30),
            "#ff7f50": (255, 127, 80),
            "#4b0082": (75, 0, 130),
            "#e6e6fa": (230, 230, 250),
            "#ee82ee": (238, 130, 238),
            "#ffdab9": (255, 218, 185),
            "#ffd700": (255, 215, 0),
            "#dda0dd": (221, 160, 221),
            "#da70d6": (218, 112, 214),
            "#dc143c": (220, 20, 60),
            "#7fff00": (127, 255, 0),
            "#ff6347": (255, 99, 71),
            "#fa8072": (250, 128, 114),
            "#fff5ee": (255, 245, 238),
            "#de5d83": (222, 93, 131),
            "#f5deb3": (245, 222, 179),
            "#f0e68c": (240, 230, 140),
            "#bdfcc9": (189, 252, 201),
            "#228b22": (34, 139, 34),
            "#fff700": (255, 247, 0),
            "#eaeaea": (234, 234, 234),
            "#ccccff": (204, 204, 255),
            "#c2b280": (194, 178, 128),
            "#50c878": (80, 200, 120),
            "#80461b": (128, 70, 27),
            "#ff2400": (255, 36, 0),
            "#36454f": (54, 69, 79),
            "#ff637d": (255, 99, 123),
            "#ff8100": (255, 129, 0),
            "#7f0029": (127, 0, 41)
        }

        self.color_names_hex:dict[str, str] = {
            "red": "#ff0000",
            "green": "#00ff00",
            "blue": "#0000ff",
            "yellow": "#ffff00",
            "cyan": "#00ffff",
            "magenta": "#ff00ff",
            "black": "#000000",
            "white": "#ffffff",
            "gray": "#a9a9a9",
            "grey": "#a9a9a9",
            "silver": "#c0c0c0",
            "maroon": "#800000",
            "olive": "#808000",
            "lime": "#00ff00",
            "aqua": "#00ffff",
            "fuchsia": "#ff00ff",
            "teal": "#008080",
            "navy": "#000080",
            "purple": "#800080",
            "pink": "#ffc0cb",
            "brown": "#8b4513",
            "beige": "#f5f5dc",
            "chocolate": "#d2691e",
            "coral": "#ff7f50",
            "indigo": "#4b0082",
            "lavender": "#e6e6fa",
            "violet": "#ee82ee",
            "peach": "#ffdab9",
            "gold": "#ffd700",
            "plum": "#dda0dd",
            "orchid": "#da70d6",
            "crimson": "#dc143c",
            "chartreuse": "#7fff00",
            "tomato": "#ff6347",
            "salmon": "#fa8072",
            "seashell": "#fff5ee",
            "blush": "#de5d83",
            "wheat": "#f5deb3",
            "khaki": "#f0e68c",
            "mint": "#bdfcc9",
            "forest": "#228b22",
            "lemon": "#fff700",
            "pearl": "#eaeaea",
            "periwinkle": "#ccccff",
            "sand": "#c2b280",
            "emerald": "#50c878",
            "russet": "#80461b",
            "scarlet": "#ff2400",
            "charcoal": "#36454f",
            "rose": "#ff637d",
            "tangerine": "#ff8100",
            "bordeaux": "#7f0029"
        }

        self.color = self.ColorToRGB(color.lower())
        self.alpha = alpha

    def RGB(self, r:int, g:int, b:int) -> tuple:
        """Sets the color to the specified RGB values

        Args:
            r (int): Red.
            g (int): Green.
            b (int): Blue.

        Returns:
            tuple: Just a copy of the entered data in the form of a tuple.
        """
        self.color = (r, g, b)
        return (r, g, b)

    def ColorToRGB(self, color:str) -> tuple:
        """Converts the specified color into its specified RGB values IF IT IS PRESENT IN THE DICTIONARY.

        Args:
            color (str): The color name.

        Returns:
            tuple: The RGB values.
        """
        if not color.lower() in self.color_names_rgb.keys():
            return None
        
        return self.color_names_rgb[color.lower()]
    
    def HexToRGB(self, hex_code:str) -> tuple:
        """Converts Hex code to RGB values IF IT IS PRESENT IN THE DICTIONARY.

        Args:
            hex_code (str): The hexadecimal code.

        Returns:
            tuple: The RGB values.
        """
        if not hex_code.lower() in self.hex_codes_rgb.keys():
            return None
        
        return self.hex_codes_rgb[hex_code.lower()]
    
    def ColorToHex(self, color:str) -> tuple:
        """Converts a color name to a hexadecimal code IF IT IS PRESENT IN THE DICTIONARY.

        Args:
            color (str): The color name.

        Returns:
            tuple: The Hexadecimal value.
        """
        if not color.lower() in self.color_names_hex.keys():
            return None
        
        return self.color_names_hex[color.lower()]

    def to_rgb_code(self):
        """Converts the RGB tuple to an ANSI escape code."""
        return f"\033[38;2;{self.color[0]};{self.color[1]};{self.color[2]}m"

    def to_background_code(self):
        """Converts the RGB tuple to an ANSI background escape code."""
        return f"\033[48;2;{self.color[0]};{self.color[1]};{self.color[2]}m"

class PHeader:
    """Generates a PHardware Header.
    
    Args:
        version (int, optional): The version of the program/File. Defaults to 1.
        flags (int, optional): The flags.. Defaults to 0.
        magicCode (Union[str, bytes], optional): The magic code in ASCII or string. Defaults to "PHardwareITK".
    """
    def __init__(self, version:int=1, flags:int=0, magicCode:Union[str, bytes]="PHardwareITK") -> None:
        """Initialize the PHeader.

        Args:
            version (int, optional): The version of the program/File. Defaults to 1.
            flags (int, optional): The flags.. Defaults to 0.
            magicCode (Union[str, bytes], optional): The magic code in ASCII or string. Defaults to "PHardwareITK".
        """
        self.version = version
        self.flags = flags
        self.magicCode = magicCode
        self.checksum = None

        if isinstance(magicCode, str):
            self.magicCode = magicCode.encode('ascii')
            magicCode = magicCode.encode('ascii')
        elif isinstance(magicCode, bytes):
            try:
                magicCode.decode('ascii')
            except UnicodeDecodeError:
                return None
    
    def calculateChecksum(self, Header:bytes) -> str:
        """Calculate checksum of Header.

        Args:
            Header (bytes): The header.

        Returns:
            str: The checksum.
        """
        import hashlib
        return hashlib.sha256(Header).hexdigest()
    
    def Pack(self) -> Any:
        """Packs the specified data into a single PHeader.

        
        The structure includes:

        
        - Magic number (4 bytes)

        - Version (4 bytes)

        - Flags (4 bytes)

        - Placeholder for checksum (32 bytes as hexadecimal string)


        Returns:
            Any: The header.
        """
        import struct

        headerData = struct.pack('>4sIB', self.magicCode, self.version, self.flags)

        checkSum = self.calculateChecksum(headerData)

        return headerData + bytes.fromhex(checkSum)
    
    def UnPack(self, data:str) -> Any:
        """Upacks the binary header into header fields.
        """
        import struct
        header = data[:-64]
        checksum = data[-64:]

        magic, version, flags = struct.unpack('>4sIB', header)
        
        self.magic = magic
        self.version = version
        self.flags = flags
        self.checksum = checksum.decode()
        
        # Validate checksum
        expected_checksum = self.calculateChecksum(header)
        if self.checksum != expected_checksum:
            raise ValueError(f"Checksum mismatch: expected {expected_checksum}, but got {self.checksum}")
        
        return self
    
    def WriteToFile(self, filename):
        """
        Write the packed header data to a binary file.
        """
        header_data = self.Pack()
        with open(filename, 'wb') as f:
            f.write(header_data)

    def ReadFromFile(self, filename):
        """
        Read the header from a binary file and unpack the fields.
        """
        if not os.path.exists(filename):
            raise FileNotFoundError(f"{filename} does not exist.")
        
        with open(filename, 'rb') as f:
            data = f.read()
        
        return self.UnPack(data)

class PBin:
    """
    PBin (PHardware Binary) is a custom binary serialization format designed to serialize various data types into a
    compact binary representation. The format includes strings, integers, floats, and custom flags.

    Attributes:
        data (bytes): The binary data representing the serialized object.
    """
    def __init__(self, header:Optional[PHeader]) -> None:
        """
        Initializes the PBin object with an empty data field.
        """
        self.data:bytes = b""

        self.header:PHeader = header if header else PHeader()

    def Serialize(self, data:Union[str, bytes, int, dict, float, list, bool], CB_flag:Optional[bool]=None) -> bytes:
        """
        Serializes a string, integer, float, and a custom flag into a binary format.

        Args:
            data (Union[str, bytes (UTF-8), int, dict (CONVERTED TO STR), float, list, bool]): The data to serialize, in UTF-8 format.
            CB_flag (Optional[bool], optional): A flag indicating whether a custom byte follows. Defaults to None

        Returns:
            bytes: The serialized binary data.
        """
        import random

        if isinstance(data, list):
            data = "".join(data)
        elif isinstance(data, bytes):
            data = data.decode()
        elif isinstance(data, int) or (isinstance(data, float) or (isinstance(data, dict) or (isinstance(data, bool)))):
            data = str(data)

        binary_data = bytearray()

        # Serialize the header first using the PHeader class
        header_data = self.header.Pack()
        binary_data.extend(header_data)

        # 1. String serialization (length-prefixed UTF-8 string)
        encoded_string: bytes = data.encode('utf-8')
        binary_data.append(len(encoded_string))  # Length of the string as a single byte.
        binary_data.extend(encoded_string)  # Actual string data

        # 2. Custom flag section (1 byte flag to indicate custom data)
        if CB_flag:
            binary_data.append(0xFF)  # Custom flag byte to indicate extra data
            binary_data.append(random.randint(0, 255))  # Random byte between 0 and 255 for custom data

        # Return the complete binary data.
        self.data = bytes(binary_data)
        return self.data
    
    def Deserialize(self, binary_data:bytes) -> tuple[str, bool, Optional[int]]:
        """
        Deserializes the provided binary data into its original components.

        Args:
            binary_data (bytes): The binary data to deserialize.

        Returns:
            tuple: A tuple containing the deserialized values in the following order:
                - string_data (str): The deserialized string.
                - CB_flag (bool): Flag indicating whether custom data exists.
                - CB_data (Optional[int]): Custom data byte if present.
        """
        import io
        import struct

        # Unpack the header first
        header_data = binary_data[:self.header.Pack().__len__()]
        self.header.UnPack(header_data)

        # Continue with the rest of the deserialization
        stream = io.BytesIO(binary_data[len(header_data):])

        # 1. Deserialize string (length-prefixed)
        string_length: int = ord(stream.read(1))  # Read the length of the string (1 byte)
        string_data: str = stream.read(string_length).decode('utf-8')  # Read the string

        # 2. Check if there's a custom data flag
        custom_flag: bool = False
        custom_data: int = None
        if stream.read(1) == b'\xFF':
            custom_flag = True
            custom_data = ord(stream.read(1))  # Read the custom data byte

        return (string_data, custom_flag, custom_data)
    
    @property
    def BinaryData(self):
        return self.data

def ToBytes(data:Optional[Union[str, dict, int, bool]]) -> Optional[bytes]:
    """Encode the given data.

    Args:
        data (Optional[Union[str, dict, int, bool]]): The data.

    Returns:
        Optional[bytes]: Encoded data.
    """
    if not data:
        return None
    data = str(data)
    return data.encode()

def FromBytes(data:Optional[bytes]) -> Optional[str]:
    """Decode the given data.

    Args:
        data (Optional[bytes]): The data.

    Returns:
        Optional[str]: Decoded data.
    """
    if not data:
        return None
    return data.decode()

class TextFont:
    """Just a Class for Font.
    """
    def __init__(self, font:str="Arial", font_size:int=11, font_color:Color=Color(), font_background_color:Color=Color("black"), Italic:bool=False, Underline:bool=False,
                 Bold:bool=False, XtraBold:bool=False, StrikeThrough:bool=False) -> None:
        """Initialize for TextFont class.

        Args:
            font (str, optional): Font. Defaults to "Arial".
            font_size (int, optional): The size of the font. Defaults to 11.
            font_color (Color, optional): The color of the font. Defaults to Color().
            font_background_color (Color, optional): The bg color of the font. Defaults to Color("black").
            Italic (bool, optional): Font Italic. Defaults to False.
            Underline (bool, optional): Underlined Font. Defaults to False.
            Bold (bool, optional): Bold Font. Defaults to False.
            XtraBold (bool, optional): More Bolder Font. Defaults to False.
            StrikeThrough (bool, optional): StikeThrough Font. Defaults to False.
        """
        
        if Bold and XtraBold:
            Bold = False

        self.font = font
        self.size = font_size
        self.color = font_color
        self.background_color = font_background_color
        self.italic = Italic
        self.underline = Underline
        self.bold = Bold
        self.xtraBold = XtraBold
        self.strike_through = StrikeThrough

    def to_font_code(self):
        """Generate the font style escape codes."""
        codes = []
        if self.bold:
            codes.append("1")  # Bold
        if self.xtraBold:
            codes.append("8")  # Extra Bold (if supported by terminal)
        if self.italic:
            codes.append("3")  # Italic
        if self.underline:
            codes.append("4")  # Underline
        if self.strike_through:
            codes.append("9")  # Strike-through
        
        if codes:
            return f"\033[{';'.join(codes)}m"
        return ""
    
    def to_reset_code(self):
        """Reset the font styles."""
        return "\033[0m"
    
    def to_font_size_code(self):
        """Returns the Escape codes for Font Size. NOTE: Does't work with all OS/Terminals. Due to Font size not being supported universally.
        OS like -> MacOS, XTerm Based, GNOME (Not via Escape codes), Windows (By changing data in a windows file), etc do support it.
        
        Supported Platforms for this function -> MacOS (ITerm2), Xterm Based (Includes Konsole).
        """

        if self.size == 11:
            return None # Default size

        if "iTerm" in os.environ.get("TERM_PROGRAM", ""):
            return f"\033]50;SetProfile=font-size={self.size}\a" #ITerm2
        elif "xterm" in os.environ.get("TERM", "") or "konsole" in os.environ.get("TERM", ""):
            return f"ESC]50;SetProfile=font-size={self.size}" # Xterm & Konsole
        else:
            return None # Other OS/Terminals

class PWidget:
    def __init__(self, x:int=0, y:int=0, width:int=5, height:int=5, radius:int=0, bgColor:Color=Color(), textWrap:bool=False, textFont:TextFont=TextFont(), text:str="") -> None:
        """Initialize function for PWidget class.

        Args:
            x (int, optional): X coordinate. Defaults to 0.
            y (int, optional): Y coordinate. Defaults to 0.
            width (int, optional): Width of widget. Defaults to 5.
            height (int, optional): Height of widget. Defaults to 5.
            radius (int, optional): Radius of widget (If required). Defaults to 0.
            bgColor (Color, optional): Background color of widget. Defaults to Color().
            textWrap (bool, optional): Text In Widget?. Defaults to False.
            textFont (TextFont, optional): The Text Font (if required). Defaults to TextFont().
            text (str, optional): The text. (if required). Defaults to "".
        """
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.radius = radius
        self.bgColor = bgColor
        self.textWrap = textWrap
        self.textFont = textFont
        self.text = text

    def Contains(self, x:int, y:int) -> bool:
        """Check if a point (x, y) is within the widget's boundaries."""
        return self.x <= x <= self.x + self.width and self.y <= y <= self.y + self.height

    def Draw(self, window: Any):
        """Draw the widget onto the window. Should be overridden by subclasses."""
        raise NotImplementedError("draw() method must be implemented by the subclass.")

    def Change(self, x:Optional[int]=None, y:Optional[int]=None, width:Optional[int]=None, height:Optional[int]=None, radius:Optional[int]=None, bgColor:Optional[Color]=None, textWrap:Optional[bool]=None, textFont:Optional[TextFont]=None, text:Optional[str]=None) -> None:
        """Change the details.

        Args:
            x (Optional[int], optional): X. Defaults to None.
            y (Optional[int], optional): Y. Defaults to None.
            width (Optional[int], optional): Width. Defaults to None.
            height (Optional[int], optional): Height. Defaults to None.
            radius (Optional[int], optional): Radius. Defaults to None.
            bgColor (Optional[Color], optional): Background Color. Defaults to None.
            textWrap (Optional[bool], optional): Text Wrap?. Defaults to None.
            textFont (TextFont, optional): Text Font. Defaults to None.
            text (Optional[str], optional): Text. Defaults to None.
        """
        if x:
            self.x = x
        
        if y:
            self.y = y

        if width:
            self.width = width

        if height:
            self.height = height

        if radius:
            self.radius = radius

        if bgColor:
            self.bgColor = bgColor

        if textWrap != None:
            self.textWrap = textWrap

        if text:
            self.text = text

        if textFont:
            self.textFont = textFont

class PIcon:
    """PIcon (PHardware Icon) class.
    """
    def __init__(self, iconPath: Optional[str] = "") -> None:
        self.iconPath:Optional[str] = iconPath
        self.BiconPath:Optional[bytes] = self.ToBytes(iconPath)

    def LoadImageSDL2(self) -> Any:
        """Loads the image for SDL2.

        Returns:
            Any: Loaded image.
        """
        import sdl2
        return sdl2.SDL_LoadBMP(self.BiconPath)
    
    def SetIconSDL2(self, window:Any) -> None:
        """Changes the icon in the specified window. SDL2.

        Args:
            window (Any): The window.
        """
        import sdl2
        
        if self.BiconPath:
            sdl2.SDL_SetWindowIcon(window, self.LoadImageSDL2())
        else:
            return None

    def Free(self) -> None:
        """Frees the image.
        """
        import sdl2
        if self.LoadImageSDL2():
            sdl2.SDL_FreeSurface(self.LoadImageSDL2())

    def ToBytes(self, data:Optional[Union[str, dict, int, bool]]) -> Optional[bytes]:
        """Encode the given data.

        Args:
            data (Optional[Union[str, dict, int, bool]]): The data.

        Returns:
            Optional[bytes]: Encoded data.
        """
        if not data:
            return None
        data = str(data)
        return data.encode()

    def FromBytes(self, data:Optional[bytes]) -> Optional[str]:
        """Decode the given data.

        Args:
            data (Optional[bytes]): The data.

        Returns:
            Optional[str]: Decoded data.
        """
        if not data:
            return None
        return data.decode()


import platform
import sdl2

import sys
import os

PHardwareITK = len(sys.path) - 1
PHardwareITK_P = os.path.join(os.path.dirname(__file__), "..", "..")

if not sys.path[PHardwareITK] == PHardwareITK_P:
    sys.path.append(PHardwareITK_P)

from phardwareitk.Extensions import HyperOut
from phardwareitk import Extensions

from typing import *

def get_platform() -> str:
    return platform.system().lower()

def SafeExitSDL(msg:Optional[Union[str, int, dict, bool, list, tuple]], DestroyFunc:object, DestroyFuncParams:tuple) -> None:
    """Safe Exit for SDL.

    Args:
        msg (Optional[Union[str, int, dict, bool, list, tuple]]): Message.
        DestroyFunc (object): The DestroyFunc in gui_sdl.py.
        DestroyFuncParams (tuple): The DestroyFunc Parameters in gui_sdl.py.
    """
    DestroyFunc(*DestroyFuncParams)
    if msg:
        msg = str(msg)
        if msg != "":
            HyperOut.printH(msg, FontEnabled=True, Font=Extensions.TextFont(font_color=Extensions.Color("red")))
    
    sdl2.SDL_Quit()

#Check Platform
OS:str = get_platform()
Linux:bool = False
Windows:bool = False
Unix:bool = False
Unknown_os:bool = True

QUIT = 256 #SDL_QUIT

if OS == "windows":
    Windows = True
    Unknown_os = False
elif OS == "linux":
    Linux = True
    Unknown_os = False
elif OS == "unix":
    Unix = True
    Unknown_os = False
else:
    Unknown_os = True

# Constants
WM_DESTROY:int = 2
WM_PAINT:int = 15